# -*- coding: utf-8 -*-
"""
Module to handle graphs.
Base structures :
- undirected, unweighted graph: a list of neighbours index (list of numpy array).
"""

import logging

import numpy as np

import pyhrf

from pyhrf.tools import cartesian
from pyhrf.boldsynth.spatialconfig import lattice_indexes
from pyhrf.boldsynth.spatialconfig import mask_to_coords
from pyhrf.ndarray import expand_array_in_mask


logger = logging.getLogger(__name__)


try:
    from pygraph.algorithms.searching import breadth_first_search as bfs_pygraph
    # raise ImportError() # for test
except ImportError, e:
    if 0 and pyhrf.__usemode__ == pyhrf.DEVEL:
        logger.warning('---------------------------------------------------')
        logger.warning('Warning: pygraph is not available. Maybe it is not installed.')
        logger.warning('You may try "easy_install python-graph-core" or the package'
                       '"python-pygraph" may be available from ubuntu reps.')
        logger.warning('---------------------------------------------------')
    pygraph_available = False
else:
    pygraph_available = True


def graph_is_sane(g, toroidal=False):
    """ Check the structure of graph 'g', which is a list of neighbours index.
    Return True if check is ok, else False.
    -> every nid in g[id] should verify id in g[nid]
    -> any neighbours list must have unique elements
    -> no isolated node
    """
    nn = len(g[0])
    for i, nl in enumerate(g):
        # if len(nl) == 0:
        #    print 'orphan node'
        #    return False
        if toroidal and len(nl) != nn:
            logger.info('not consistent with toroidal geometry')
            return False
        if np.unique(nl).size != nl.size:
            logger.info('duplicate neighbours, nl: %s', str(nl))
            return False
        for k in nl:
            # print 'k:', k
            # print 'nl:', nl
            if i not in g[k]:
                logger.info(
                    '%d is neighbour of %d but converse is not true', i, k)
                return False
    return True


def graph_from_mesh(polygonList):
    """Return the list of neighbours indexes for each position, from a list of
    polygons. Each polygon is a triplet.

    """

    indexSet = set()
    for l in polygonList:
        indexSet.update(l)
    nbPositions = len(indexSet)

    logger.debug('indexSet:')
    logger.debug(indexSet)

    neighbourSets = {}
    for triangle in polygonList:
        for idx in triangle:
            if not neighbourSets.has_key(idx):
                neighbourSets[idx] = set()
            neighbourSets[idx].update(triangle)

    # idx<->set --> idx<->list
    neighbourLists = np.empty(nbPositions, dtype=object)
    for idx in indexSet:
        neighbourSets[idx].remove(idx)
        neighbourLists[idx] = np.array(list(neighbourSets[idx]), dtype=int)

    return neighbourLists


def graph_nb_cliques(graph):
    return sum([len(x) for x in graph]) / 2.


def center_mask_at_v01(mask, pos, shape):
    new = range(len(mask))
    toKeep = set(range(len(mask[0])))
    for i, (d, p, s) in enumerate(zip(mask, pos, shape)):
        comp = d + p
        new[i] = comp
        toKeep.difference_update(np.where(comp < 0)[0])
        toKeep.difference_update(np.where(comp >= s)[0])
    toKeep = list(toKeep)
    # print 'toKeep:', toKeep
    # print 'new:', new
    return tuple([n[toKeep] for n in new])


def center_mask_at(mask, pos, indexes, toroidal=False):
    ndim = len(mask)
    # print 'ndim:', ndim
    new = range(ndim)
    # print 'new:', new
    toKeep = np.ones(len(mask[0]), dtype=bool)
    # print 'toKeep:', toKeep
    for ic in xrange(ndim):
        # print 'ic:', ic
        # print 'pos[ic]:', pos[ic]
        comp = mask[ic] + pos[ic]
        # print 'comp:', comp
        if not toroidal:
            insiders = np.bitwise_and(comp >= 0, comp < indexes.shape[ic])
            np.bitwise_and(toKeep, insiders, toKeep)
        else:
            comp[np.where(comp < 0)] = indexes.shape[ic] - 1
            comp[np.where(comp >= indexes.shape[ic])] = 0
            # print 'indexes.shape[ic]:', indexes.shape[ic]
            # print 'comp after modif:', comp
        new[ic] = comp
        # print 'new[ic]:', new[ic]
    #m = tuple( [n[toKeep] for n in new] )
    #np.bitwise_and(toKeep, indexes[m]!=-1, toKeep)
    # print 'toKeep:', toKeep
    # print 'new:', new
    return tuple([n[toKeep] for n in new])


def center_mask_at_v02(mask, pos, shape):
    new = [[] for n in xrange(len(mask))]
    # print ' mask:', mask
    # print ' pos:', pos
    # print 'new:', new
    ndim = len(mask)
    for i in xrange(len(mask[0])):
        ncoords = range(ndim)
        valid = True
        for ic in xrange(ndim):
            ncoord = mask[ic][i] + pos[ic]
            if ncoord < 0 or ncoord >= shape[ic]:
                valid = False
                break
            else:
                ncoords[ic] = ncoord
        if valid:
            for ic in xrange(ndim):
                new[ic].append(ncoords[ic])
    return tuple(new)


# kerMask3D_6n = np.array( [ [1,0,0]  ,  [0,1,0] , [0,0,1],
#                           [-1,0,0] , [0,-1,0] , [0,0,-1] ], dtype=int )
kerMask3D_6n = (np.array([1, 0, 0, -1, 0, 0], dtype=int),
                np.array([0, 1, 0, 0, -1, 0], dtype=int),
                np.array([0, 0, 1, 0, 0, -1], dtype=int))

#kerMask2D_4n = np.array( [ [-1,0], [1,0], [0,1], [0,-1] ], dtype=int )
kerMask2D_4n = (np.array([-1, 1, 0, 0], dtype=int),
                np.array([0, 0, 1, -1], dtype=int))

kerMask2D_8n = (np.array([-1, 1, 0, 0, -1, 1, -1, 1], dtype=int),
                np.array([0, 0, 1, -1, -1, 1, 1, -1], dtype=int))


def graph_pool_indexes(g):
    nodeMap = dict([(v, iv) for iv, v in enumerate(nodes)])
    for nl in subg:
        for i, n in enumerate(nl):
            nl[i] = nodeMap[n]


def flatten_and_graph(data, mask=None, kerMask=None, depth=1,
                      toroidal=False):

    if mask is None:
        mask = data > 0

    fdata = data[np.where(mask)]
    g = graph_from_lattice(mask, kerMask, depth, toroidal)
    return fdata, g

#---------------------------   Lotfi   -----------------------------#


def graph_from_lattice3D(mask, kerMask=None, depth=1, toroidal=False):
    """
    Creates a graph from a n-dimensional lattice
    'mask' define valid positions to build the graph over.
    'kerMask' is numpy array mask (tuple of arrays) which defines the
    neighbourhood system, ie the relative positions of neighbours for a given
    position in the lattice.
    """
    # print 'size:', latticeIndexes.size
    if kerMask is not None:
        assert mask.ndim == len(kerMask)
    else:
        # Full neighbourhood, ie 8 in 2D, 26 in 3D ...
        ndim = mask.ndim
        neighbourCoords = list(cartesian(*[[0, -1, 1]] * ndim))[1:]
        kerMask = tuple(np.array(neighbourCoords, dtype=int).transpose())

    # loop over valid positions:

    positions = np.array(np.where(mask >= 0)).transpose()
    # print "----------------------------"
    # print np.prod(mask.shape)
    # print positions.shape
    # print "----------------------------"
    # raw_input('')
    #positions = mask_to_coords(mask)
    latticeIndexes = lattice_indexes(mask)
    # print positions
    # print positions.shape
    # Build lists of closest neighbours:
    closestNeighbours = np.empty(len(positions), dtype=object)
    for idx, pos in enumerate(positions):
        # print 'idx :', idx, '- pos:', pos
        # O(n) :
        m = center_mask_at(kerMask, pos, latticeIndexes, toroidal)
        closestNeighbours[idx] = latticeIndexes[m][latticeIndexes[m] >= 0]

    if depth == 1:
        return closestNeighbours

    neighboursSets = copyModule.deepcopy(closestNeighbours)

    for idx in xrange(nbPositions):
        # Starting positions = closest neighbours of current position
        neighboursToTreat = closestNeighbours[idx]
        visited = set([idx])
        # O(max( sum(lattice shape), kerMask size))*O(kerMask size)*O(d)
        for d in xrange(depth - 1):
            newNeighboursToTreat = set()
            # In the worst case, neighbours to treat are on the perimeter
            # of the lattice, say O(sum(lattice shape))
            # OR
            # If the lattice is kind of flat, the worst case is
            # O(size of kerMask)
            # So complexity is :
            # O(max( sum(lattice shape), kerMask size)) * O(kerMask size)
            for n in neighboursToTreat:
                # O(kerMask size)
                newNeighboursToTreat.update(closestNeighbours[n])
            # Remember what we just saw:
            visited.update(neighboursToTreat)
            # Update neighbours of current position and remove those
            # already seen:
            newNeighboursToTreat.difference_update(visited)
            neighboursSets[idx].update(newNeighboursToTreat)
            # Define new starting positions for next loop:
            neighboursToTreat = newNeighboursToTreat

    return neighboursSets

#--------------------------------   Lotfi   -------------------------#


def graph_to_sparse_matrix(graph):
    """
    Creates a connectivity sparse matrix from the adjacency graph
    (list of neighbors list)
    """
    from scipy.sparse.coo import coo_matrix
    n_vox = len(graph)

    ij = [[], []]
    for i in xrange(n_vox):
        ij[0] += [i] * len(graph[i])
        ij[1] += list(graph[i])

    return coo_matrix((np.ones(len(ij[0]), dtype=int), ij))


def graph_from_lattice(mask, kerMask=None, depth=1, toroidal=False):
    """
    Creates a graph from a n-dimensional lattice
    'mask' define valid positions to build the graph over.
    'kerMask' is numpy array mask (tuple of arrays) which defines the
    neighbourhood system, ie the relative positions of neighbours for a given
    position in the lattice.
    """
    # print 'size:', latticeIndexes.size
    if kerMask is not None:
        assert mask.ndim == len(kerMask)
    else:
        # Full neighbourhood, ie 8 in 2D, 26 in 3D ...
        ndim = mask.ndim
        neighbourCoords = list(cartesian(*[[0, -1, 1]] * ndim))[1:]
        kerMask = tuple(np.array(neighbourCoords, dtype=int).transpose())

    # loop over valid positions:
    # print mask.shape
    positions = mask_to_coords(mask)
    latticeIndexes = lattice_indexes(mask)
    # print positions
    # print positions.shape
    # Build lists of closest neighbours:
    closestNeighbours = np.empty(len(positions), dtype=object)
    for idx, pos in enumerate(positions):
        # print 'idx :', idx, '- pos:', pos
        # O(n) :
        m = center_mask_at(kerMask, pos, latticeIndexes, toroidal)
        # print 'm:', m
        closestNeighbours[idx] = latticeIndexes[m][latticeIndexes[m] >= 0]

    if depth == 1:
        return closestNeighbours

    neighboursSets = copyModule.deepcopy(closestNeighbours)

    for idx in xrange(nbPositions):
        # Starting positions = closest neighbours of current position
        neighboursToTreat = closestNeighbours[idx]
        visited = set([idx])
        # O(max( sum(lattice shape), kerMask size))*O(kerMask size)*O(d)
        for d in xrange(depth - 1):
            newNeighboursToTreat = set()
            # In the worst case, neighbours to treat are on the perimeter
            # of the lattice, say O(sum(lattice shape))
            # OR
            # If the lattice is kind of flat, the worst case is
            # O(size of kerMask)
            # So complexity is :
            # O(max( sum(lattice shape), kerMask size)) * O(kerMask size)
            for n in neighboursToTreat:
                # O(kerMask size)
                newNeighboursToTreat.update(closestNeighbours[n])
            # Remember what we just saw:
            visited.update(neighboursToTreat)
            # Update neighbours of current position and remove those
            # already seen:
            newNeighboursToTreat.difference_update(visited)
            neighboursSets[idx].update(newNeighboursToTreat)
            # Define new starting positions for next loop:
            neighboursToTreat = newNeighboursToTreat

    return neighboursSets


def breadth_first_search(graph, root=0, visitable=None):
    """Traverses a graph in breadth-first order.

    The first argument should be the tree root; visitable should be an
    iterable with all searchable nodes;
    """

    if visitable is None:
        visitable = range(len(graph))  # all nodes are visitable

    # makes a shallow copy, makes it a collection, removes duplicates
    unvisited = list(set(visitable))

    queue = []
    if root in unvisited:
        unvisited.remove(root)
        queue.append(root)

    order = []
    while len(queue) > 0:
        node = queue.pop(0)
        order.append(node)

        for child in graph[node]:
            if child in unvisited:
                unvisited.remove(child)
                queue.append(child)

    return None, order  # return None to be compliant with pygraph which also
    # return the spanning tree


def graph_pygraph(g):
    from pygraph.classes.graph import graph
    gr = graph()
    gr.add_nodes(range(len(g)))
    for inode, nl in enumerate(g):
        for ineighbour in nl:
            if ineighbour > inode:
                gr.add_edge((inode, ineighbour))
    return gr


def connected_components_iter(g):
    if pygraph_available:
        g = graph_pygraph(g)
        bfs = bfs_pygraph
    else:
        logger.warning('Warning: pygraph not available ... fall back to slow BFS function')
        bfs = breadth_first_search

    visited = np.zeros(len(g), dtype=bool)
    root = 0
    while not visited.all():
        root = np.where(visited == False)[0][0]
        # print 'root:', root
        st, order = bfs(g, root=root)
        visited[order] = True
        yield sorted(order)


def connected_components(g):

    if pygraph_available:
        g = graph_pygraph(g)
        bfs = bfs_pygraph
    else:
        logger.warning('Warning: pygraph not available ... fall back to slow BFS function')
        bfs = breadth_first_search

    if 1:
        visited = np.zeros(len(g), dtype=bool)
        components = []
        root = 0
        while not visited.all():
            root = np.where(visited == False)[0][0]
            st, order = bfs(g, root=root)
            visited[order] = True
            components.append(sorted(order))
        return components
    else:
        from pygraph.algorithms.accessibility import connected_components as cc
        return cc(g)

if pygraph_available:
    def connected_components_labeled(g, labels):

        gr = graph_pygraph(g)

        counts = np.bincount(labels)
        to_visit = set(range(len(gr)))

        class label_filter(object):
            def __init__(self, l, labels):
                self.l = l
                self.labels = labels

            def configure(self, gr, s):
                pass

            def __call__(self, node, parent):
                return self.labels[node] == self.l

        components = dict([(i, []) for i in np.unique(labels)])
        while len(to_visit) > 0:
            root = to_visit.pop()
            label = labels[root]
            st, o = bfs_pygraph(gr, root, label_filter(label, labels))
            components[label].append(o)
            to_visit.difference_update(o)

        return components


def split_mask_into_cc_iter(mask, min_size=0, kerMask=None):
    """ Return an iterator over all connected components (CC) within input mask. CC which are smaller than `min_size`
    are discarded. `kerMask` defines the connectivity, e.g., kerMask3D_6n for 6-neighbours in 3D.

    Examples
    --------

    .. code::

        vol = np.array( [[1,1,0,1,1],
                        [1,1,0,1,1],
                        [0,0,0,0,0],
                        [1,0,1,1,0],
                        [0,0,1,1,0]], dtype=int )
        for cc in split_mask_into_cc_iter(vol):
            print cc


    Should output:

    .. code::

        np.array( [[1,1,0,0,0],
                   [1,1,0,0,0],
                   [0,0,0,0,0],
                   [0,0,0,0,0],
                   [0,0,0,0,0]]
        np.array( [[0,0,0,1,1],
                   [0,0,0,1,1],
                   [0,0,0,0,0],
                   [0,0,0,0,0],
                   [0,0,0,0,0]]

    """
    assert isinstance(min_size, int)
    flat_mask = mask[np.where(mask)]
    # print 'flat_mask:', flat_mask.shape
    g = graph_from_lattice(mask, kerMask)
    for cc in connected_components_iter(g):
        # print 'cc:'
        # print cc
        if len(cc) >= min_size:
            flat_mask[:] = 0
            flat_mask[cc] = 1
            # print expand_array_in_mask(flat_mask, mask)
            yield expand_array_in_mask(flat_mask, mask)


def bfs_set_label(g, root, data, value, radius):
    from pygraph.algorithms.filters.radius import radius as Radius
    from pygraph.algorithms.searching import breadth_first_search
    gr = graph_pygraph(g)
    filt_radius = Radius(radius)
    _, o = breadth_first_search(gr, root, filt_radius)
    data[o] = value


def bfs_sub_graph(g, root, radius):
    from pygraph.algorithms.filters.radius import radius as Radius
    from pygraph.algorithms.searching import breadth_first_search
    gr = graph_pygraph(g)
    filt_radius = Radius(radius)
    _, o = breadth_first_search(gr, root, filt_radius)
    to_keep = sorted(o)
    new_indexes = np.zeros(len(g))
    new_indexes[to_keep] = range(len(to_keep))
    subg = g[to_keep]
    for nl in subg:
        for i in xrange(len(nl)):
            nl[i] = new_indexes[nl[i]]
    return subg, to_keep


def sub_graph(graph, nodes):
    sNodeIds = set(nodes)
    # Filter the graph to only keep given nodes:
    subg = np.array([np.array(list(sNodeIds.intersection(nl)))
                     for nl in graph[nodes]], dtype=object)
    # Fix indexes to be in the new referential:
    nodeMap = dict([(v, iv) for iv, v in enumerate(nodes)])
    for nl in subg:
        for i, n in enumerate(nl):
            nl[i] = nodeMap[n]
    return subg, nodeMap


def parcels_to_graphs(parcellation, kerMask, toKeep=None, toDiscard=None):
    """Compute graphs for each parcel in parcels. A graph is simply defined as a list of neighbour indexes.

    Parameters
    ----------
    parcellation:
        is a n-ary numpy array.
    kerMask:
        defines the connectivity
    toKeep
    toDiscard

    Returns
    -------
    a dictionary mapping a roi ID to its graph

    """
    # determine the set of parcels to work on:
    parcelIds = np.unique(parcellation)  # everything
    if toKeep is not None:
        assert set(toKeep).issubset(set(parcelIds))
        parcelIds = toKeep
    if toDiscard is not None:
        #assert set(toDiscard).issubset(set(parcelIds))
        parcelIds = set(parcelIds).difference(toDiscard)
    # print 'parcelIds:', parcelIdsx

    parcelGraphs = {}
    parcelNodeMap = {}
    # loop over parcels
    for pid in parcelIds:
        parcelGraphs[pid] = graph_from_lattice(parcellation == pid,
                                               kerMask=kerMask)
        # print 'pg:', parcelGraphs[pid]
        assert graph_is_sane(parcelGraphs[pid])

    return parcelGraphs



# -*- coding: utf-8 -*-

import unittest
import shutil
import tempfile
import os
import os.path as op

import numpy as np

import pyhrf
import pyhrf.tools._io as pio

from pyhrf import get_data_file_name
from pyhrf.ndarray import MRI3Daxes, MRI4Daxes
from pyhrf.ndarray import xndarray


class RxCopyTest(unittest.TestCase):

    def setUp(self,):
        self.tmp_dir = pyhrf.get_tmp_path()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def _create_tmp_files(self, fns):
        for fn in [op.join(self.tmp_dir, fn) for fn in fns]:
            d = op.dirname(fn)
            if not op.exists(d):
                os.makedirs(d)
            open(fn, 'a').close()

    def assert_file_exists(self, fn, test_exists=True):
        if test_exists and not op.exists(fn):
            raise Exception('File %s does not exist' % fn)
        elif not test_exists and op.exists(fn):
            raise Exception('File %s exists' % fn)

    def test_basic(self):
        self._create_tmp_files([op.join('./raw_data', f)
                                for f in ['AC0832_anat.nii', 'AC0832_asl.nii',
                                          'AC0832_bold.nii', 'PK0612_asl.nii',
                                          'PK0612_bold.nii', 'dummy.nii']])
        src = '(?P<subject>[A-Z]{2}[0-9]{4})_(?P<modality>[a-zA-Z]+).nii'
        src_folder = op.join(self.tmp_dir, 'raw_data')
        dest_folder = (self.tmp_dir, 'export', '{subject}', '{modality}')
        dest_basename = 'data.nii'
        pio.rx_copy(src, src_folder, dest_basename, dest_folder)

        for fn in [op.join(self.tmp_dir, 'export', f)
                   for f in ['AC0832/bold/data.nii',
                             'AC0832/anat/data.nii',
                             'AC0832/asl/data.nii',
                             'PK0612/bold/data.nii',
                             'PK0612/asl/data.nii']]:
            self.assert_file_exists(fn)

    def test_advanced(self):
        self._create_tmp_files([op.join('./raw_data', f)
                                for f in ['ASL mt_TG_PASL_s004a001.nii',
                                          'ASL mt_TG_PASL_s008a001.nii',
                                          'ASL mt_PK_PASL_s064a001.nii',
                                          'ASL mt_PK_PASL_s003a001.nii']])
        src = 'ASL mt_(?P<subject>[A-Z]{2})_(?P<modality>[a-zA-Z]+)_'\
              's(?P<session>[0-9]{3})a[0-9]{3}.nii'
        src_folder = op.join(self.tmp_dir, 'raw_data')
        dest_folder = (self.tmp_dir, 'export', '{subject}', '{modality}')
        dest_basename = 'ASL_session_{session}.nii'
        pio.rx_copy(src, src_folder, dest_basename, dest_folder)

        for fn in [op.join(self.tmp_dir, 'export', f)
                   for f in ['TG/PASL/ASL_session_004.nii',
                             'TG/PASL/ASL_session_008.nii',
                             'PK/PASL/ASL_session_064.nii',
                             'PK/PASL/ASL_session_003.nii']]:
            self.assert_file_exists(fn)

    def test_with_subfolders(self):
        self._create_tmp_files([op.join('./raw_data/ASL', f)
                                for f in ['mt_TG_PASL_s004a001.nii',
                                          'mt_TG_PASL_s008a001.nii',
                                          'mt_PK_PASL_s064a001.nii',
                                          'mt_PK_PASL_s003a001.nii']])

        self._create_tmp_files([op.join('./raw_data/BOLD', f)
                                for f in ['mt_TG_BOLDepi_s003a001.nii',
                                          'mt_TG_BOLDepi_s005a001.nii',
                                          'mt_PK_BOLDepi_s022a001.nii',
                                          'mt_PK_BOLDepi_s007a001.nii']])

        src = '(?P<modality_folder>[a-zA-Z]+)/mt_(?P<subject>[A-Z]{2})_' \
              '(?P<modality>[a-zA-Z]+)_'\
              's(?P<session>[0-9]{3})a[0-9]{3}.nii'
        src_folder = op.join(self.tmp_dir, 'raw_data')
        dest_folder = (self.tmp_dir, 'export', '{subject}', '{modality}')
        dest_basename = '{modality}_session_{session}.nii'
        pio.rx_copy(src, src_folder, dest_basename, dest_folder)

        for fn in [op.join(self.tmp_dir, 'export', f)
                   for f in ['TG/PASL/PASL_session_004.nii',
                             'TG/PASL/PASL_session_008.nii',
                             'TG/BOLDepi/BOLDepi_session_003.nii',
                             'TG/BOLDepi/BOLDepi_session_005.nii',
                             'PK/PASL/PASL_session_064.nii',
                             'PK/PASL/PASL_session_003.nii',
                             'PK/BOLDepi/BOLDepi_session_022.nii',
                             'PK/BOLDepi/BOLDepi_session_007.nii']]:
            self.assert_file_exists(fn)

    def test_missing_tags_dest_folder(self):
        self._create_tmp_files(['AK98_T1_s01.nii'])
        src_folder = self.tmp_dir
        src = '(?P<subject>[A-Z]{2}[0-9]{2})_(?P<modality>[a-zA-Z0-9]+)'
        dest_folder = (self.tmp_dir, 'export', '{study}', '{modality}',
                       '{session}')
        dest_basename = '{subject}.nii'
        self.assertRaisesRegexp(pio.MissingTagError,
                                "Tags in dest_folder not defined in src: "
                                "study, session", pio.rx_copy,
                                src, src_folder, dest_basename, dest_folder)

    def test_missing_tags_dest_basename(self):
        self._create_tmp_files(['AK98_T1_s01.nii'])
        src_folder = self.tmp_dir
        src = '[A-Z]{2}[0-9]{2}_(?P<modality>[a-zA-Z0-9]+)'
        dest_folder = (self.tmp_dir, 'export', '{modality}')
        dest_basename = '{subject}_{session}.nii'
        self.assertRaisesRegexp(pio.MissingTagError,
                                "Tags in dest_basename not defined in src: "
                                "(subject, session)|(session, subject)",
                                pio.rx_copy, src, src_folder,
                                dest_basename, dest_folder)

    def test_dry(self):
        self._create_tmp_files(['AK98_T1_s01.nii'])
        src_folder = self.tmp_dir
        src = '[A-Z]{2}[0-9]{2}_(?P<modality>[a-zA-Z0-9]+)'
        dest_folder = (self.tmp_dir, 'export', '{modality}')
        dest_basename = 'data.nii'
        pio.rx_copy(src, src_folder, dest_basename, dest_folder, dry=True)
        fn = op.join(self.tmp_dir, 'export', 'T1', 'data.nii')
        if op.exists(fn):
            raise Exception('File %s should not exist' % fn)

    def test_duplicates_targets(self):
        self._create_tmp_files(['AK98_T1_s01.nii', 'AK98_T1_s02.nii'])
        src_folder = self.tmp_dir
        src = '[A-Z]{2}[0-9]{2}_(?P<modality>[a-zA-Z0-9]+).*nii'
        dest_folder = (self.tmp_dir, 'export', '{modality}')
        dest_basename = 'data.nii'
        error_msg = r'Copy is not injective, the following copy ' \
                    'operations have the same destination:\n'   \
                    '.*AK98_T1_s01\.nii\n.*AK98_T1_s02\.nii\n'      \
                    '-> .*export/T1/data.nii'
        self.assertRaisesRegexp(pio.DuplicateTargetError, error_msg,
                                pio.rx_copy, src, src_folder,
                                dest_basename, dest_folder)

    def test_replacement(self):
        self._create_tmp_files([op.join('./raw_data', f)
                                for f in ['ASL mt_TG_PASL_s004a001.nii',
                                          'ASL mt_TG_T1_s008a001.nii',
                                          'ASL mt_PK_PASL_s064a001.nii',
                                          'ASL mt_PK_T1_s003a001.nii']])
        src = 'ASL mt_(?P<subject>[A-Z]{2})_(?P<modality>[a-zA-Z0-9]+)_'\
              's(?P<session>[0-9]{3})a[0-9]{3}.nii'
        src_folder = op.join(self.tmp_dir, 'raw_data')
        dest_folder = (self.tmp_dir, 'export', '{subject}', '{modality}')
        dest_basename = '{modality}_session_{session}.nii'
        pio.rx_copy(src, src_folder, dest_basename, dest_folder,
                    replacements=[('T1', 'anat'), ('PASL', 'aslf')])
        for fn in [op.join(self.tmp_dir, 'export', f)
                   for f in ['TG/aslf/aslf_session_004.nii',
                             'TG/anat/anat_session_008.nii',
                             'PK/aslf/aslf_session_064.nii',
                             'PK/anat/anat_session_003.nii']]:
            self.assert_file_exists(fn)

    def test_callback(self):
        def filter_odd_session(s, d):
            if (int(d[-5]) % 2) != 0:
                return None
            else:
                return d

        self._create_tmp_files([op.join('./raw_data', f)
                                for f in ['ASL mt_TG_PASL_s004a001.nii',
                                          'ASL mt_TG_T1_s008a001.nii',
                                          'ASL mt_PK_PASL_s064a001.nii',
                                          'ASL mt_PK_T1_s003a001.nii']])
        src = 'ASL mt_(?P<subject>[A-Z]{2})_(?P<modality>[a-zA-Z0-9]+)_'\
              's(?P<session>[0-9]{3})a[0-9]{3}.nii'
        src_folder = op.join(self.tmp_dir, 'raw_data')
        dest_folder = (self.tmp_dir, 'export', '{subject}', '{modality}')
        dest_basename = '{modality}_session_{session}.nii'
        pio.rx_copy(src, src_folder, dest_basename, dest_folder,
                    replacements=[('T1', 'anat'), ('PASL', 'aslf')],
                    callback=filter_odd_session)
        for fn in [op.join(self.tmp_dir, 'export', f)
                   for f in ['TG/aslf/aslf_session_004.nii',
                             'TG/anat/anat_session_008.nii',
                             'PK/aslf/aslf_session_064.nii']]:
            self.assert_file_exists(fn)

        self.assert_file_exists('PK/anat/anat_session_003.nii', False)


class NiftiTest(unittest.TestCase):

    def setUp(self):
        tmpDir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                  dir=pyhrf.cfg['global']['tmp_path'])
        self.tmp_dir = tmpDir
        self.clean_tmp = True

    def tearDown(self):
        if self.clean_tmp:
            shutil.rmtree(self.tmp_dir)

    def test_process_history_extension(self):
        nii_fn = pyhrf.get_data_file_name(
            'real_data_vol_4_regions_mask.nii.gz')

        nii_fn_out = op.join(self.tmp_dir, 'proc_ext_test.nii')
        input_pname = 'dummy_proc_test'
        input_pparams = {'my_param': 5.5, 'input_file': '/home/blh'}

        pio.append_process_info(nii_fn, input_pname, input_pparams,
                                img_output_fn=nii_fn_out)

        i2, (aff, header) = pio.read_volume(nii_fn_out)

        reloaded_pinfo = pio.get_process_info(nii_fn_out)
        self.assertNotEqual(reloaded_pinfo, None)
        self.assertEqual(reloaded_pinfo[0]['process_name'], input_pname)
        self.assertEqual(reloaded_pinfo[0]['process_inputs'], input_pparams)
        self.assertEqual(reloaded_pinfo[0]['process_version'], None)
        self.assertEqual(reloaded_pinfo[0]['process_id'], None)


class DataLoadTest(unittest.TestCase):

    def test_paradigm_csv(self):
        pfn = get_data_file_name('paradigm_loc_av.csv')
        o, d = pio.load_paradigm_from_csv(pfn)
        if 0:
            print 'onsets:'
            print o
            print 'durations:'
            print d

    def test_paradigm_csv2(self):
        pfn = get_data_file_name('paradigm_loc_av.csv')
        o, d = pio.load_paradigm_from_csv(pfn, " ")
        if 0:
            print 'onsets:'
            print o
            print 'durations:'
            print d

    def test_paradigm_csv3(self):
        pfn = get_data_file_name('paradigm_loc_av_comma.csv')
        o, d = pio.load_paradigm_from_csv(pfn)

    def test_paradigm_csv4(self):
        pfn = get_data_file_name('paradigm_loc_av_comma.csv')
        o, d = pio.load_paradigm_from_csv(pfn, ",")

    def test_frmi_vol(self):
        """ Test volumic data loading
        """
        boldFn = get_data_file_name('subj0_bold_session0.nii.gz')
        roiMaskFn = get_data_file_name('subj0_parcellation.nii.gz')
        g, b, ss, m, h = pio.load_fmri_vol_data([boldFn, boldFn], roiMaskFn)
        if 0:
            print len(g), g[1]
            print b[1].shape
            print ss
            print m.shape, np.unique(m)
            print h


class xndarrayIOTest(unittest.TestCase):

    def setUp(self):
        self.cub0 = xndarray(np.random.rand(10, 10))
        self.cub3DVol = xndarray(np.random.rand(10, 10, 10),
                                 axes_names=MRI3Daxes)
        d4D = np.zeros((2, 2, 2, 3))
        for c in xrange(3):
            d4D[:, :, :, c] = np.ones((2, 2, 2)) * (c - 2)

        self.cub4DVol = xndarray(d4D, axes_names=['condition'] + MRI3Daxes)

        self.cub4DTimeVol = xndarray(np.random.rand(100, 10, 10, 10),
                                     axes_names=['time'] + MRI3Daxes)
        self.cubNDVol = xndarray(np.random.rand(10, 2, 2, 2, 3),
                                 axes_names=['time'] +
                                 MRI3Daxes + ['condition'],
                                 axes_domains={'condition': ['audio', 'video', 'na']})

        self.tmp_dir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                        dir=pyhrf.cfg['global']['tmp_path'])

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_save_nii_3D(self):
        fn = op.join(self.tmp_dir, 'test3D.nii')
        self.cub3DVol.save(fn)

    def test_save_nii_4D(self):
        fn = op.join(self.tmp_dir, 'test4D.nii')
        self.cub4DTimeVol.save(fn)

    def test_save_nii_multi(self):
        c = self.cubNDVol.reorient(MRI4Daxes + ['condition'])
        c.save(op.join(self.tmp_dir, './testND.nii'))


class FileHandlingTest(unittest.TestCase):

    def setUp(self):

        self.tmp_dir = pyhrf.get_tmp_path()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_split_ext(self):
        bfn = pyhrf.get_data_file_name('subj0_bold_session0.nii.gz')
        pio.split_ext_safe(bfn)

    def test_split4DVol(self):
        s = 'subj0_bold_session0.nii.gz'
        bfn = pyhrf.get_data_file_name(s)
        bold_files = pio.split4DVol(bfn, output_dir=self.tmp_dir)
        i, meta = pio.read_volume(bold_files[0])

        for bf in bold_files:
            os.remove(bf)


class GiftiTest(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                        dir=pyhrf.cfg['global']['tmp_path'])

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_read_tex_gii_label(self):
        tex_fn = 'real_data_surf_tiny_parcellation.gii'
        tex_fn = pyhrf.get_data_file_name(tex_fn)
        t, tgii = pio.read_texture(tex_fn)

    def test_read_default_real_data_tiny(self):
        mesh_file = pyhrf.get_data_file_name('real_data_surf_tiny_mesh.gii')
        bold_file = pyhrf.get_data_file_name('real_data_surf_tiny_bold.gii')
        fn = 'real_data_surf_tiny_parcellation.gii'
        parcel_file = pyhrf.get_data_file_name(fn)

        cor, tri, mesh_gii = pio.read_mesh(mesh_file)
        bold, bold_gii = pio.read_texture(bold_file)
        parcellation, parcel_gii = pio.read_texture(parcel_file)

    def test_load_fmri_surf_data(self):
        """ Test surfacic data loading
        """
        mesh_file = pyhrf.get_data_file_name('real_data_surf_tiny_mesh.gii')
        bold_file = pyhrf.get_data_file_name('real_data_surf_tiny_bold.gii')
        fn = 'real_data_surf_tiny_parcellation.gii'
        parcel_file = pyhrf.get_data_file_name(fn)

        g, b, ss, m, el = pio.load_fmri_surf_data([bold_file, bold_file],
                                                  mesh_file,
                                                  parcel_file)
        assert len(g) == len(np.unique(m))

    def test_write_tex_gii_labels(self):
        labels = np.random.randint(0, 2, 10)
        tex_fn = op.join(self.tmp_dir, 'labels.gii')
        pio.write_texture(labels, tex_fn)
        t, tgii = pio.read_texture(tex_fn)
        assert t.dtype == labels.dtype
        assert (t == labels).all()

    def test_write_tex_gii_float(self):
        values = np.random.randn(10)
        tex_fn = op.join(self.tmp_dir, 'float_values.gii')
        pio.write_texture(values, tex_fn)
        t, tgii = pio.read_texture(tex_fn)
        assert t.dtype == values.dtype
        assert np.allclose(t, values)

    def test_write_tex_gii_time_series(self):
        values = np.random.randn(120, 10).astype(np.float32)
        tex_fn = op.join(self.tmp_dir, 'time_series.gii')
        pio.write_texture(values, tex_fn, intent='time series')
        t, tgii = pio.read_texture(tex_fn)
        assert t.dtype == values.dtype
        assert np.allclose(t, values)

    def test_write_tex_gii_2D_float(self):
        values = np.random.randn(2, 10).astype(np.float32)
        tex_fn = op.join(self.tmp_dir, 'floats_2d.gii')
        pio.write_texture(values, tex_fn)
        t, tgii = pio.read_texture(tex_fn)
        assert t.dtype == values.dtype
        assert np.allclose(t, values)


class SPMIOTest(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = pyhrf.get_tmp_path()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def _test_load_regnames(self, spm_ver):
        spm_file = op.join(self.tmp_dir, 'SPM_v{}.mat'.format(spm_ver))
        pio._zip.gunzip(pyhrf.get_data_file_name('SPM_v%d.mat.gz' % spm_ver),
                        outFileName=spm_file)
        expected = ['Sn(1) audio*bf(1)', 'Sn(1) video*bf(1)',
                    'Sn(2) audio*bf(1)', 'Sn(2) video*bf(1)',
                    'Sn(1) constant', 'Sn(2) constant']
        self.assertEqual(pio.spmio.load_regnames(spm_file), expected)

    def test_load_regnames_SPM8(self):
        self._test_load_regnames(8)

    def test_load_regnames_SPM5(self):
        self._test_load_regnames(5)

    def test_load_regnames_SPM12(self):
        self._test_load_regnames(12)



# -*- coding: utf-8 -*-


from numpy import *
#from scipy.stats import truncnorm, erfc
from scipy.stats import truncnorm
from scipy.special import erfc
from scipy.special import erf

class RandomGenerator():
    """B
    Abstract class to ensure the definition of the function generate.
    """
    def __call__(self, size):
        return self.generate(size)

    def generate(self, size):
        raise NotImplementedError

class GaussianGenerator(RandomGenerator):
    """
    Class encapsulating the gaussian random generator of numpy
    """
    def __init__(self, mean=0.0, var=1.0):
        self.mean = mean
        self.std = var**.5
        self.type = 'gaussian'
    
    def generate(self, size):
        return random.normal(self.mean, self.std, size)

class GammaGenerator(RandomGenerator):
    """
    Class encapsulating the gamma random generator of numpy
    """
    def __init__(self, mean=1.0, var=1.0):
        self.a = mean**2/(var+0.)
        self.b = var/(mean+0.)
        self.type = 'gamma'

    def generate(self, size):
        return random.gamma(self.a, self.b, size)

class BetaGenerator(RandomGenerator):
    """
    Class encapsulating the beta random generator of numpy
    """
    def __init__(self, mean=.5, var=0.1):
        frac = mean*(1. - mean)/var - 1
        self.a = mean * frac
        self.b = frac -self.a
        self.type = 'beta'

    def generate(self, size):
        return random.beta(self.a, self.b, size)

class LogNormalGenerator(RandomGenerator):
    """
    Class encapsulating the log normal generator of numpy
    """
    def __init__(self, meanLogN=1.0, varLogN=1.0):
        self.mean = log(meanLogN) - .5 * log( 1.+ varLogN/meanLogN**2) 
        self.std = sqrt( log( 1. + varLogN/meanLogN**2 ) )
        self.type = 'log-normal'

    def generate(self, size):
        return random.lognormal(self.mean, self.std, size)
    
class UniformGenerator(RandomGenerator):
    """
    Class encapsulating the random generator
    """
    def __init__(self, minV=0., maxV=1.0):
        self.minV = minV 
        self.maxV =maxV
        self.widthInterval = maxV -minV
        self.type = 'uniform'
    
    def generate(self, size):
        return self.widthInterval * random.rand(size) - + self.minV

class ZeroGenerator(RandomGenerator):
    """
    Class encapsulating the null distribution !!!!!!!!!
    """
    def __init__(self):
        self.type = 'zero-valued'
    
    def generate(self, size):
        return zeros(size, dtype=float)

class IndependentMixtureLaw:
    """
    Class handling the generation of values following an indenpendent mixture
    law. Requires the prior generator of label values. 
    """
    def __init__(self, states, generators):
        """
        Initialise an IndependentMixtureLaw instance i.e. a set of values
        generated by 'generators' according to the labels given in 'states' 
        """
        self.states = states
        nbc = self.states.getNbClasses()
        self.nbClasses = nbc
        self.generators = generators
        self.data = zeros(self.states.getSize(), dtype=float)
        self.generate()
        
    def generate(self):
        """
        Generate realisations of the mixture law.
        """
        for ic in xrange(self.states.getNbClasses()):
            labels = self.states.getFieldValues()
            m = where(labels==ic)
            cn = self.states.getClassName(ic)
            if cn != None:
                self.data[m] = self.generators[cn].generate(len(m[0]))
            #else a generator is not defined for this class, leave zeros



def rpnorm(n,m,s):
    """
    Random numbers from the positive normal distribution.
    rpnorm(n,m,s) is a vector of length n with random entries, generated
    from a positive normal distribution with mean m and standard
    deviation s.

    Original matlab code from:
    (c) Vincent Mazet, 06/2005
    Centre de Recherche en Automatique de Nancy, France
    vincent.mazet@cran.uhp-nancy.fr
    
    Reference:
    V. Mazet, D. Brie, J. Idier, 'Simulation of Positive Normal Variables
    using several Proposal Distributions', IEEE Workshop Statistical
    Signal Processing 2005, july 17-20 2005, Bordeaux, France.

    Adapted by Thomas VINCENT:
    thomas.vincent@cea.fr
    """

    if s<0:  Exception('Standard deviation must be positive.')
    if n<=0: Exception('n is wrong.')

    x = []

    nn = n

    #Intersections
    ca  = 1.136717791056118
    mA = (1-ca**2)/ca*s
    mC = s * sqrt(pi/2)
    
    while len(x)<nn:    
        if m < mA:     # 4. Exponential distribution
            a = (-m + sqrt(m**2+4*s**2)) / 2 / s**2
            z = -log(1-random.rand(n))/a
            rho = exp( -(z-m)**2/2/s**2 - a*(m-z+a*s**2/2) )

        elif m <= 0:   # 3. Normal distribution truncated at the mean
                       #  equality because 3 is faster to compute than the 2
            z = abs(random.randn(n))*s + m
            rho = (z>=0)

        elif m < mC:   # 2. Normal distribution coupled with the uniform one
            r = (random.rand(n) < m/(m+sqrt(pi/2)*s))
            u = random.rand(n)*m
            g = abs(random.randn(n)*s) + m
            z = r*u + (1-r)*g
            rho = r*exp(-(z-m)**2/2/s**2) + (1-r)*ones(n)

        else:          # 1. Normal distribution
            z = random.randn(n)*s + m
            rho = (z>=0)


        # Accept/reject the propositions
        reject = (random.rand(n) > rho)
        z = z[bitwise_not(reject)]
        if len(z)>0:
            x.extend(z)
        n = n-len(z)

    return array(x)



def truncRandn(size, mu=0., sigma=1., a=0., b=inf):

    if b == inf:
        return rpnorm(size, mu, sigmaC) + a
    else:
        tn = truncnorm(a, b, mu, sigma)
        return tn.rvs(size)


import numpy as np
from numpy.random import rand, randn

def gm_sample(means, variances, props, n=1):

    assert np.isscalar(n)
    assert len(means) == len(variances)
    assert len(props) == len(means)

    if not isinstance(means, np.ndarray):
        means = np.array(means)

    if not isinstance(variances, np.ndarray):
        variances = np.array(variances)


    labels = (rand(n)[:,np.newaxis] > np.cumsum(props)[np.newaxis,:]).sum(1)

    return randn(n) * variances[labels]**.5 + means[labels]



# -*- coding: utf-8 -*-


import numpy as np
from scipy.integrate import trapz
from pyhrf.tools import resampleToGrid


def compute_T_Pvalue(betas, stds_beta, mask_file, null_hyp=True):
    '''
    Compute Tvalues statistic and Pvalue based upon estimates
    and their standard deviation
    beta and std_beta for all voxels
    beta: shape (nb_vox, 1)
    std: shape (1)
    Assume null hypothesis if null_hyp is True
    '''
    from pyhrf.ndarray import xndarray

    import sys
    sys.path.append("/home/i2bm/BrainVisa/source/pyhrf/pyhrf-free/trunk/script/WIP/Scripts_IRMf_Adultes_Solv/Scripts_divers_utiles/Scripts_utiles/")
    from Functions_fit import Permutation_test, stat_mean, stat_Tvalue, stat_Wilcoxon

    mask = xndarray.load(mask_file).data #to save P and Tval on a map

    BvalC = xndarray(betas, axes_names=['sagittal', 'coronal', 'axial'])
    Betasval = BvalC.flatten(mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data

    Stdsval = stds_beta

    Tval = xndarray(Betasval/Stdsval, axes_names=['position']).data

    nb_vox = Betasval.shape[0]
    nb_reg = betas.shape[1]
    dof = nb_vox - nb_reg #degrees of freedom for STudent distribution
    assert dof>0

    Probas=np.zeros(Betasval.shape)
    for i in xrange(nb_vox):
        if null_hyp:
            #STudent distribution
            from scipy.stats import t
            fmix = lambda x: t.pdf(x, dof)
        else:
            fmix = lambda t:  1/np.sqrt(2*np.pi*Stdsval[i]**2)*np.exp(- (t - Betasval[i])**2 / (2*Stdsval[i]**2) )
        Probas[i] = quad(fmix, Tval[i], float('inf'))[0]

    Tvalues_ = xndarray(Tval, axes_names=['position'])
    Pvalues_ = xndarray(Probas, axes_names=['position'])
    Tvalues = Tvalues_.expand(mask, 'position', ['sagittal','coronal','axial'])
    Pvalues = Pvalues_.expand(mask, 'position', ['sagittal','coronal','axial'])

    #Computation of Pvalue using permutations
    #not possible to do this actually...it was used for group level stats
    #Pvalue_t = np.zeros(Betasval.shape)
    #for i in xrange(nb_vox):
        #Pvalue_t[i] = Permutation_test(Betasval[i], n_permutations=10000, \
                    #stat = stat_Tvalue, two_tailed=False, plot_histo=False)

    return Tvalues.data, Pvalues.data

def compute_roc_labels_scikit(e_labels, true_labels):
    from scikits.learn.metrics import roc_curve
    from scikits.learn.metrics import auc as compute_auc
    from scipy import interp
    auc = []
    sensData = []
    specData = []
    grid_fpr = np.linspace(0,1,100)
    for j in xrange(true_labels.shape[0]):
        evalues = e_labels[j,:]
        tvalues = true_labels[j,:]
        # print 'evalues:', evalues.shape
        # print evalues
        # print 'tvalues:', tvalues.shape
        # print tvalues
        fpr, tpr, thresh = roc_curve(tvalues, evalues)
        # print 'fpr:'
        # print fpr
        # print 'tpr:'
        # print tpr
        # print 'thresh:'
        # print thresh
        auc.append(compute_auc(fpr, tpr))
        # print 'auc:', auc[-1]
        sensData.append(interp(grid_fpr, fpr, tpr))
        specData.append(grid_fpr)
        sensData[-1][0] = 0.
        # print 'sensData:'
        # print sensData[-1]
        # print 'specData:'
        # print specData[-1]
    sensData = np.array(sensData)
    specData = np.array(specData)
    auc = np.array(auc)
    return sensData, specData, auc

def threshold_labels(labels, thresh=None, act_class=1):
#def threshold_labels(labels, thresh, act_class=1):
    """
    Threshold input labels which are assumed being of shape
    (nb classes, nb conds, nb vox).
    If thresh is None then take the argmax over classes.
    Else use it on labels for activating class (act_class), suitable
    for the 2class case only.
    """
    finalLabels = np.zeros(labels.shape[1:], dtype=np.int32)
    for j in xrange(finalLabels.shape[0]):
        if thresh is None:
            finalLabels[j,:] = np.argmax(labels[:,j,:], axis=0)
        else:
            actVox = np.where(labels[act_class,j,:]>=thresh)
            finalLabels[j,actVox] = act_class
    return finalLabels

def mark_wrong_labels(labels, true_labels, lab_ca=1, lab_ci=0,false_pos=2,
                      false_neg=3):

    new_labels = labels.copy()
    for j in xrange(labels.shape[0]):
        #print 'labels :'
        #print labels[j,:]
        #print 'trueLabels '
        #print self.trueLabels[j,:]
        el = (labels[j,:] == lab_ca)
        tl = (true_labels[j,:] == lab_ca)
        nel = np.bitwise_not(el)
        ntl = np.bitwise_not(tl)
        new_labels[j, np.bitwise_and(el,tl)] = lab_ca
        new_labels[j, np.bitwise_and(nel,ntl)] = lab_ci
        new_labels[j, np.bitwise_and(el,ntl)] = false_pos
        new_labels[j, np.bitwise_and(nel,tl)] = false_neg
        #print '-> marked :'
        #print labels[j,:]

    return new_labels

def compute_roc_labels(mlabels, true_labels, dthres=0.005, lab_ca=1, lab_ci=0,
                false_pos=2, false_neg=3):
    thresholds = np.arange(0,1/dthres) * dthres
    nconds = true_labels.shape[0]
    oneMinusSpecificity = np.zeros((nconds, len(thresholds)))
    sensitivity = np.zeros((nconds, len(thresholds)))
    for it,thres in enumerate(thresholds):
	#print 'thres:',thres
        labs = threshold_labels(mlabels,thres)
        tlabels = labs.copy()
        labs = mark_wrong_labels(labs,true_labels,lab_ca, lab_ci, false_pos, false_neg)
        for cond in xrange(nconds):
            if 0 and cond == 1:
                print "**cond %d **" %cond
                print 'thresh:', thres
                print 'mlabels (activ class):'
                print mlabels[1,cond,:80]
                print 'thresholded labels:'
                print tlabels[cond,:80]
                print 'simulated labels:'
                print true_labels[cond,:80]
                print 'marked labels:'
                print labs[cond,:80]
            counts = np.bincount(labs[cond,:])
            nbTrueNeg = counts[0]
            nbTruePos = counts[1] if len(counts)>1 else 0

            fp = false_pos
            nbFalsePos = counts[fp] if len(counts)>fp else 0

            fn = false_neg
            nbFalseNeg = counts[fn] if len(counts)>fn else 0

            #if cond == 1 or cond == 2:
	      #print 'cond ', cond
	      #print 'nbTrueNeg=',nbTrueNeg
	      #print 'nbTruePos=',nbTruePos
	      #print 'nbFalsePos=',nbFalsePos
	      #print 'nbFalseNeg=',nbFalseNeg

            if 0 and cond == 1:
                print 'TN :', nbTrueNeg
                print 'TP :', nbTruePos
                print 'FP :', nbFalsePos
                print 'FN :', nbFalseNeg
            if nbTruePos == 0:
                sensitivity[cond,it] = 0
            else:
                sensitivity[cond,it] = nbTruePos /                    \
                                       (nbTruePos+nbFalseNeg+0.0)
            spec = nbTrueNeg/(nbTrueNeg+nbFalsePos+0.0)
            oneMinusSpecificity[cond,it] = 1-spec
            if 0 and cond == 1:
                print '-> se = ', sensitivity[cond, it]
                print '-> 1-sp = ', oneMinusSpecificity[cond,it]

    if 1:
        spGrid = np.arange(0.,1.,0.0005)
        omspec = np.zeros((nconds, len(spGrid)))
        sens = np.zeros((nconds, len(spGrid)))
        for cond in xrange(nconds):
            order = np.argsort(oneMinusSpecificity[cond,:])
            if oneMinusSpecificity[cond,order][0] != 0.:
                osp = np.concatenate(([0.],oneMinusSpecificity[cond,order]))
                se = np.concatenate(([0.],sensitivity[cond,order]))
            else:
                osp = oneMinusSpecificity[cond,order]
                se = sensitivity[cond,order]

            if osp[-1] != 1.:
                osp = np.concatenate((osp,[1.]))
                se = np.concatenate((se,[1.]))

            sens[cond,:] = resampleToGrid(osp, se, spGrid)
            omspec[cond, :] = spGrid
            if 0 and cond == 1:
                print '-> (se,1-spec) :'
                print zip(se,osp)
                print '-> se resampled :'
                print sens[cond,:]
                print 'spec grid :'
                print spGrid


    else:
        sens = sensitivity
        omspec = oneMinusSpecificity

    auc = np.array([trapz(sens[j,:], omspec[j,:])
                    for j in xrange(nconds)])

    #Computing the area under ROC curve (with John Walkenbach formula) (SAME AS auc)
    #area_under_ROC_curve = np.zeros(nconds, dtype=float)
    #for cond in xrange(nconds):
      #for i in xrange(len(spGrid)-1):
	#if sens[cond,i]*sens[cond,i+1] >= 0.:
	  #area_under_ROC_curve[cond] += ((sens[cond,i+1] + sens[cond,i])/2.) * (omspec[cond,i+1] - omspec[cond,i])
	#else:
	  #area_under_ROC_curve[cond] += ((sens[cond,i+1]**2 + sens[cond,i]**2)/((sens[cond,i+1] - sens[cond,i])/2.)) * (omspec[cond,i+1] - omspec[cond,i])

    return sens, omspec, auc #, area_under_ROC_curve


def cumFreq(data, thres=None):

    nBins = data.ptp()/(3.5*data.std()/(len(data)**.33)) # Scott's choice
    #nBins = np.around(np.log2(len(data))+1) # Sturges' formula

    h,b = np.histogram(data, bins=nBins,normed=True)
    ccBins = (b[:-1]+b[1:])/2. # class center approximation
##    print 'ccBins :'
##    print ccBins

    if thres != None:
        # Find bin where contrast <= thres :
        infT = (ccBins <= thres)+0
##        print 'infT :'
##        print infT
        if infT.sum() == len(ccBins):
            iThres = -1 # take all
        else:
            iThres = infT.tolist().index(0)
    else:
        iThres = -1 # take all
##    print 'iThres :', iThres
    # cut ccH at given threshold :
    h = h[:iThres]
    ccBins = ccBins[:iThres]

##    print 'H :'
##    print h
##    print 'ccBins :'
##    print ccBins


    # Append zeros at begining and end to finalize approximation:
    h = np.concatenate(([0],h,[0]))
    ccBins = np.concatenate(([b[0]-(b[1]-b[0])/2.0],
                             ccBins,
                             [b[iThres]+(b[iThres]-b[iThres-1])/2.0]))

    # Approximate the cumulative density function under given threshold:
    return np.trapz(h, ccBins)



#Functions to manipulate CDF, PDF, ...
from scipy.integrate import quad
from scipy.stats import norm

def gm_cdf(x, means, variances, props):
    """ Compute the cumulative density function of gaussian mixture,
    ie: p(x<a) = \sum_i \Nc(mean_i, variance_i)
    """
    variances[np.where(variances==0.)] = 1e-6
    return (props * norm.cdf(x, means, variances**.5)).sum(0)

def gm_mean(means, variances, props):
    if isinstance(means,list):
        means = np.array(means)

    if isinstance(props,list):
        props = np.array(props)

    return (props*means).sum()

def gm_var(means, variances, props):
    if isinstance(means,list):
        means = np.array(means)

    if isinstance(variances,list):
        variances = np.array(variances)

    if isinstance(props,list):
        props = np.array(props)

    return (props*(means**2+variances)).sum() - gm_mean(means,variances,props)**2


from scipy.stats.mstats import mquantiles
def cpt_ppm_a_mcmc(samples, alpha=0.05):
    """ Compute a Posterior Probability Map (fixed alpha) from NRL MCMC samples.
    Expected shape of 'samples': (sample, voxel)
    """
    return mquantiles(samples, prob=[1-alpha], axis=0)[0]

def cpt_ppm_g_mcmc(samples, gamma=0.):
    """ Compute a Posterior Probability Map (fixed gamma) from NRL MCMC samples.
    Expected shape of 'samples': (sample, voxel)
    """
    return (samples > gamma).mean(0)

def cpt_ppm_g_apost(means, variances, props, gamma=0.):
    """ Compute a Posterior Probability Map (fixed gamma) from posterior
    gaussian mixture components estimates.
    Expected shape of 'means', 'variances' and 'probs': (nb_classes, voxel)
    """
    return 1 - gm_cdf(gamma, means, variances, props)

def cpt_ppm_a_apost(means, variances, props, alpha=0.05):
    raise NotImplementedError('PPM with fixed alpha from posterior mixture '\
                                  'components is not implemented')

def cpt_ppm_a_norm(mean, variance, alpha=0.):
    """ Compute a Posterior Probability Map (fixed alpha) by assuming a Gaussian
    distribution.

    Parameters
    ----------
    mean : array_like
        mean value(s) of the Gaussian distribution(s)
    variance : array_like
        variance(s) of the Gaussian distribution(s)
    alpha : array_like, optional
        quantile value(s) (default=0)

    Returns
    -------
    ppm : array_like
        Posterior Probability Map evaluated at alpha
    """

    return norm.sf(alpha, mean, variance**.5)

def cpt_ppm_g_norm(mean, variance, gamma=0.95):
    """ Compute a Posterior Probability Map (fixed gamma) by assuming a Gaussian
    distribution.

    Parameters
    ----------
    mean : array_like
        mean value(s) of the Gaussian distribution(s)
    variance : array_like
        variance(s) of the Gaussian distribution(s)
    gamma : array_like, optional
        upper tail probability (default=0.95)

    Returns
    -------
    ppm : ndarray or scalar
        Posterior Probability Map corresponding to the upper tail probability gamma
    """

    return norm.isf(gamma, mean, variance**.5)


# def compute_bigaussian_ppm(threshold, mean_c1, var_c1, prop_c1,
#                            mean_c2, var_c2, prop_c2, thresh_type='value'):
#     """
#     Calculate a posterior probability map p(A>gamma)=alpha where A follows a
#     bigaussian mixture defined by the input parameters:
#     *prop_c1* Normal(*mean_c1*,*var_c1*) +
#     *prop_c2* Normal(*mean_c2*,*var_c2*).

#     *thesh_type* can be either 'value' or 'proba'.
#     If *thesh_type*='value' then the returned PPM is p(X>threshold).
#     If *thesh_type*='proba' then the returned PPM is gamma so that
#        p(X>gamma)=threshold
#     """
#     ppm = np.zeros_like(mean_c1)

#     if thresh_type == 'value':
#         def fmix(t, m1, v1, p1, m2, v2, p2):
#             val =  p1 * 1/np.sqrt(2*np.pi*v1) * \
#                 np.exp(- (t - m1)**2 / (2*v1) )  + \
#                 p2 * 1/np.sqrt(2*np.pi*v2) * \
#                 np.exp(- (t - m2)**2 / (2*v2) )
#             return val
#         for pos in xrange(len(ppm)):
#             ppm[pos] = quad(fmix, threshold, float('inf'),
#                             (mean_c1[pos], var_c1[pos], prop_c1[pos],
#                              mean_c2[pos], var_c2[pos], prop_c2[pos],)
#                             )[0]

#     elif thresh_type == 'proba':
#         raise NotImplementedError('PPM computation with proba threshold')
#     else:
#         raise Exception('Wrong thresh_type "%s", has to be '\
#                             '"value" or "proba"' %thresh_type)

#     return ppm


def acorr(x, maxlags=10, scale='var'):
    #print 'acorr ..., input x:', x.shape
    n = x.shape[0]

    maxlags = min(n-1, maxlags)

    mean = x.mean(0)
    v = ((x - mean) ** 2).sum(0) / float(n)

    #print 'c0:', v.shape

    def r(h):
        return ((x[:n - h] - mean) * (x[h:] - mean)).sum(0)

    lags = np.arange(maxlags) + 1
    ac = np.array(map(r, lags))

    #print 'ac:', ac.shape
    if scale == 'coeff':
        ac /= ac[0]
    elif scale == 'var':
        ac /= float(n) * v


    m_cst_signal = np.where(v==0.)
    for a in ac:
        a[m_cst_signal] = 1.

    return ac



    # x = detrend_mean(np.asarray(x))
    # c = np.correlate(x, x, mode=2)
    # c /= np.sqrt(np.dot(x,x)**2)


    # if maxlags is None: maxlags = Nx - 1

    # if maxlags >= Nx or maxlags < 1:
    #     raise ValueError('maxlags must be None or strictly '
    #                      'positive < %d'%Nx)

    # lags = np.arange(-maxlags,maxlags+1)
    # c = c[Nx-1-maxlags:Nx+maxlags]




# -*- coding: utf-8 -*-

import logging

import numpy as np

from pyhrf import xmlio
from pyhrf.jde.samplerbase import *
from pyhrf.jde.nrl.bigaussian import NRLSamplerWithRelVar as WNS


logger = logging.getLogger(__name__)


class W_Drift_Sampler(xmlio.XmlInitable, GibbsSamplerVariable):

    # other class attributes
    L_CI = 0
    L_CA = 1
    CLASSES = np.array([L_CI, L_CA], dtype=int)
    CLASS_NAMES = ['inactiv', 'activ']

    def __init__(self, do_sampling=True, use_true_value=False,
                 val_ini=None, pr_sigmoid_slope=1., pr_sigmoid_thresh=0.):
        """
        pr_sigmoid_slope = tau1
        pr_sigmoid_thresh = tau2
        """
        # TODO : comment
        xmlio.XmlInitable.__init__(self)

        GibbsSamplerVariable.__init__(self, 'W', valIni=val_ini,
                                      sampleFlag=do_sampling,
                                      useTrueValue=use_true_value)
        self.t1 = pr_sigmoid_slope
        self.t2 = pr_sigmoid_thresh

    def initObservables(self):
        # print 'initObservables'
        logger.info('WSampler.initObservables ...')
        GibbsSamplerVariable.initObservables(self)

    def linkToData(self, dataInput):

        self.dataInput = dataInput
        self.nbConditions = self.dataInput.nbConditions
        self.ny = self.dataInput.ny
        self.nbVoxels = self.dataInput.nbVoxels


    def checkAndSetInitValue(self, variables):


        logger.info('WSampler.checkAndSetInitW ...')
        # Generate default W if necessary :
        # if self.useTrueVal:
        #self.currentValue = np.ones(self.nbConditions, dtype=int)
        # if self.useTrueVal:
        # if self.trueW is not None:
        # TODO : take only common conditions
        #self.currentValue = self.trueW.copy()
        # else:
        #raise Exception('True W have to be used but none defined.')

        # if no initial W specified
        if self.currentValue is None or not self.sampleFlag:
            logger.info('W are not initialized '
                        '-> suppose that all conditions are relevant')
            self.currentValue = np.ones(self.nbConditions, dtype=int)

        logger.debug('init W :')
        logger.debug(self.currentValue)

    def saveCurrentValue(self, it):
        GibbsSamplerVariable.saveCurrentValue(self, it)

    def updateObsersables(self):
        GibbsSamplerVariable.updateObsersables(self)

    def saveObservables(self, it):
        GibbsSamplerVariable.saveObservables(self, it)

    def computemoyq(self, cardClassCA, nbVoxels):
        '''
        Compute mean of labels in  ROI
        '''
        moyq = np.divide(cardClassCA, (float)(nbVoxels))
        # print 'moyq=',moyq

        return moyq

    def computeProbW1(self, gj, gTgj, rb, t1, t2, mCAj, vCIj, vCAj, j, cardClassCAj):
        '''
        ProbW1 is the probability that condition is relevant
        It is a vecteur on length nbcond
        '''

        logger.info('Sampling W - cond %d ...', j)

        val = -t1 * (cardClassCAj - t2)

        tmp = (- 0.5 * cardClassCAj) * np.log(vCIj / vCAj)

        result1 = 0.
        result2 = 0.
        for i in self.voxIdx[self.L_CA][j]:
            valeur1 = (self.nrls[j, i]) ** 2 / (2. * vCIj)
            valeur2 = (self.nrls[j, i] - mCAj) ** 2 / (2. * vCAj)
            result1 += (valeur1 - valeur2)

        y = self.dataInput.varMBY
        matPl = self.samplerEngine.get_variable('drift').matPl
        StimIndSign = np.zeros((y.shape), dtype=float)
        for i in xrange(self.nbConditions):
            if i != j:
                StimIndSign += self.currentValue[i] * self.nrls[i, :] * np.tile(
                    self.varXh[:, i], (self.nbVoxels, 1)).transpose()
        ej = y - StimIndSign - matPl

        for i in xrange(self.nbVoxels):
            valeur1 = ((self.nrls[j, i]) ** 2) * gTgj
            valeur2 = 2. * self.nrls[j, i] * np.dot(ej[:, i].transpose(), gj)
            result2 += (1. / rb[i]) * (valeur1 - valeur2)

        val0 = val + tmp
        val1 = result1 - 0.5 * result2

        # print 'w =',self.currentValue

        Maximum = val0
        if (val1 > Maximum):
            Maximum = val1

        tmp2 = 1. + (np.exp(val0 - Maximum) / np.exp(val1 - Maximum))

        print 'Cond ', j, ',  val0 =', val0, ',     val1 =', val1
        # print 'cond =',j,',     v0 =',vCIj,',   v1 =',vCAj,',   m1 =',mCAj

        return 1. / tmp2

    def sampleNextInternal(self, variables):

        snrls = self.get_variable('nrl')
        self.cardClass = snrls.cardClass
        # print 'CardClass WSampler =',self.cardClass
        self.voxIdx = snrls.voxIdx
        self.nrls = snrls.currentValue

        sIMixtP = self.get_variable('mixt_params')
        var = sIMixtP.getCurrentVars()
        mean = sIMixtP.getCurrentMeans()
        sHrf = self.get_variable('hrf')
        self.varXh = sHrf.varXh
        h = sHrf.currentValue
        rb = self.get_variable('noise_var').currentValue

        varXh = sHrf.varXh
        ProbW1 = np.zeros(self.nbConditions, dtype=float)
        ProbW0 = np.zeros(self.nbConditions, dtype=float)
        self.WSamples = np.random.rand(self.nbConditions)

        gTg = np.diag(np.dot(varXh.transpose(), varXh))

        g = varXh
        cardClassCA = self.cardClass[self.L_CA, :]

        print '***************************************************************************'
        for j in xrange(self.nbConditions):
            ProbW1[j] = self.computeProbW1(g[:, j], gTg[j], rb, self.t1, self.t2, mean[
                                           self.L_CA, j], var[self.L_CI, j], var[self.L_CA, j], j, cardClassCA[j])
            ProbW0[j] = 1. - ProbW1[j]
            self.currentValue[j] = 0
            if(self.WSamples[j] <= ProbW1[j]):
                self.currentValue[j] = 1
            snrls.computeVarYTildeOptWithRelVar(self.varXh, self.currentValue)
        print 'W =', self.currentValue

    def threshold_W(self, meanW, thresh):
        print 'meanW =', meanW
        destW = np.zeros(self.nbConditions, dtype=int)
        for j in xrange(self.nbConditions):
            if meanW[j] >= thresh:
                destW[j] = 1
        return destW

    def finalizeSampling(self):

        GibbsSamplerVariable.finalizeSampling(self)
        #self.finalW = self.threshold_W(self.mean, 0.8722)
        self.finalW = self.threshold_W(self.mean, 0.5)
        print 'final w : ', self.finalW

    def getOutputs(self):

        outputs = GibbsSamplerVariable.getOutputs(self)
        cn = self.dataInput.cNames

        axes_names = ['condition']
        axes_domains = {'condition': cn}

        outputs['w_pm_thresh'] = xndarray(self.finalW, axes_names=axes_names,
                                          axes_domains=axes_domains)

        # outputs['w_pm'] = xndarray(self.mean, axes_names=axes_names,
        # axes_domains=axes_domains)

        return outputs


class WSampler(xmlio.XmlInitable, GibbsSamplerVariable):

    # old parameters specifications :
    # P_TAU1 = 'SlotOfSigmoid'
    # P_TAU2 = 'ThreshPointOfSigmoid'

    # other class attributes
    L_CI = 0
    L_CA = 1
    CLASSES = np.array([L_CI, L_CA], dtype=int)
    CLASS_NAMES = ['inactiv', 'activ']

    def __init__(self, do_sampling=True, use_true_value=False,
                 val_ini=None, pr_sigmoid_slope=1., pr_sigmoid_thresh=0.):
        """
        pr_sigmoid_slope = tau1
        pr_sigmoid_thresh = tau2
        """
        # TODO : comment
        xmlio.XmlInitable.__init__(self)

        GibbsSamplerVariable.__init__(self, 'W', valIni=val_ini,
                                      sampleFlag=do_sampling,
                                      useTrueValue=use_true_value)
        self.t1 = pr_sigmoid_slope
        self.t2 = pr_sigmoid_thresh

    def initObservables(self):
        # print 'initObservables'
        logger.info('WSampler.initObservables ...')
        GibbsSamplerVariable.initObservables(self)

    def linkToData(self, dataInput):

        self.dataInput = dataInput
        self.nbConditions = self.dataInput.nbConditions
        self.ny = self.dataInput.ny
        self.nbVoxels = self.dataInput.nbVoxels
        self.matXQ = self.dataInput.matXQ
        # if dataInput.simulData is not None:
        #self.trueW = dataInput.simulData.w
        # else:
        #self.trueW = None

    def checkAndSetInitValue(self, variables):

        # Generate default W if necessary :
        # if self.useTrueVal:
            #self.currentValue = np.ones(self.nbConditions, dtype=int)
        # if self.useTrueVal:
            # if self.trueW is not None:
                # TODO : take only common conditions
                #self.currentValue = self.trueW.copy()
            # else:
                #raise Exception('True W have to be used but none defined.')

        if self.currentValue is None:  # if no initial W specified
            logger.info('W are not initialized '
                        '-> suppose that all conditions are relevant')
            self.currentValue = np.ones(self.nbConditions, dtype=int)

        logger.debug('init W :')
        logger.debug(self.currentValue)

    def saveCurrentValue(self, it):
        GibbsSamplerVariable.saveCurrentValue(self, it)

    def updateObsersables(self):
        GibbsSamplerVariable.updateObsersables(self)

    def saveObservables(self, it):
        GibbsSamplerVariable.saveObservables(self, it)

    def computemoyq(self, cardClassCA, nbVoxels):
        '''
        Compute mean of labels in  ROI
        '''
        moyq = np.divide(cardClassCA, (float)(nbVoxels))

        return moyq

    def computeProbW1(self, Qgj, gTQgj, rb, moyqj, t1, t2, mCAj, vCIj, vCAj, j, cardClassCAj):
        '''
        ProbW1 is the probability that condition is relevant
        It is a vecteur on length nbcond
        '''
        logger.info('Sampling W - cond %d ...', j)

        val = -t1 * (cardClassCAj - t2)

        tmp = (- 0.5 * cardClassCAj) * np.log(vCIj / vCAj)

        result1 = 0.
        result2 = 0.
        for i in self.voxIdx[self.L_CA][j]:
            valeur1 = (self.nrls[j, i]) ** 2 / (2. * vCIj)
            valeur2 = (self.nrls[j, i] - mCAj) ** 2 / (2. * vCAj)
            result1 += (valeur1 - valeur2)

        y = self.dataInput.varMBY
        StimIndSign = np.zeros((y.shape), dtype=float)
        for i in xrange(self.nbConditions):
            if i != j:
                StimIndSign += self.currentValue[i] * self.nrls[i, :] * np.tile(
                    self.varXh[:, i], (self.nbVoxels, 1)).transpose()
        ej = y - StimIndSign

        for i in xrange(self.nbVoxels):
            valeur1 = ((self.nrls[j, i]) ** 2) * gTQgj
            valeur2 = 2. * self.nrls[j, i] * np.dot(ej[:, i].transpose(), Qgj)
            result2 += (1. / rb[i]) * (valeur1 - valeur2)

        val0 = val + tmp
        val1 = result1 - 0.5 * result2

        print 'w =', self.currentValue

        Maximum = val0
        if (val1 > Maximum):
            Maximum = val1

        tmp2 = 1. + np.exp(val0 - Maximum) / np.exp(val1 - Maximum)

        return 1. / tmp2

    def computeVarXhtQ(self, h, matXQ):
        for j in xrange(self.nbConditions):
            self.varXhtQ[j, :] = np.dot(h.transpose(), matXQ[j, :, :])

    def sampleNextInternal(self, variables):

        nrl = self.get_variable('nrl')
        self.cardClass = nrl.cardClass
        # print 'CardClass WSampler =',self.cardClass
        self.voxIdx = nrl.voxIdx
        self.nrls = nrl.currentValue

        self.varYtilde = nrl.varYtilde

        sIMixtP = self.get_variable('mixt_params')
        var = sIMixtP.getCurrentVars()
        mean = sIMixtP.getCurrentMeans()
        sHrf = self.get_variable('hrf')
        self.varXh = sHrf.varXh
        h = sHrf.currentValue
        rb = self.get_variable('noise_var').currentValue

        self.varXhtQ = np.zeros((self.nbConditions, self.ny), dtype=float)
        ProbW1 = np.zeros(self.nbConditions, dtype=float)
        ProbW0 = np.zeros(self.nbConditions, dtype=float)
        self.WSamples = np.random.rand(self.nbConditions)

        self.computeVarXhtQ(h, self.matXQ)
        logger.debug('varXhtQ %s :', str(self.varXhtQ.shape))
        logger.debug(self.varXhtQ)

        gTQg = np.diag(np.dot(self.varXhtQ, self.varXh))

        # Qg = QXh = (XhtQ)^t where XhtQ = h^t X^t Q and Q^t = Q
        Qg = self.varXhtQ.transpose()
        cardClassCA = self.cardClass[self.L_CA, :]
        moyq = self.computemoyq(cardClassCA, self.nbVoxels)

        for j in xrange(self.nbConditions):
            ProbW1[j] = self.computeProbW1(Qg[:, j], gTQg[j], rb, moyq[j], self.t1, self.t2, mean[
                                           self.L_CA, j], var[self.L_CI, j], var[self.L_CA, j], j, cardClassCA[j])
            ProbW0[j] = 1. - ProbW1[j]
            self.currentValue[j] = 0
            if(self.WSamples[j] <= ProbW1[j]):
                self.currentValue[j] = 1

    def threshold_W(self, meanW, thresh):
        print 'meanW =', meanW
        destW = np.zeros(self.nbConditions, dtype=int)
        for j in xrange(self.nbConditions):
            if meanW[j] >= thresh:
                destW[j] = 1
        return destW

    def finalizeSampling(self):
        #self.finalW = self.threshold_W(self.mean, 0.8722)
        GibbsSamplerVariable.finalizeSampling(self)
        self.finalW = self.threshold_W(self.mean, 0.5)
        print 'final w : ', self.finalW

    def getOutputs(self):

        outputs = GibbsSamplerVariable.getOutputs(self)
        cn = self.dataInput.cNames

        axes_names = ['condition']
        axes_domains = {'condition': cn}

        outputs['w_pm_thresh'] = xndarray(self.finalW, axes_names=axes_names,
                                          axes_domains=axes_domains)

        # outputs['w_pm'] = xndarray(self.mean, axes_names=axes_names,
        # axes_domains=axes_domains)

        return outputs



# -*- coding: utf-8 -*-
import os
import numpy as np
from pyhrf.tools import rebin

fig_orientations = {
    'axial' : ['coronal','sagittal'],
    'sagittal' : ['axial','coronal'],
    'coronal' : ['axial','sagittal'],
    }


def mix_cmap(img1, cmap1, img2, cmap2, norm1=None, norm2=None, blend_r=.5):
    import matplotlib.pyplot as plt

    assert img1.shape == img2.shape
    if norm1 is None:
        norm1 = plt.Normalize()
    cols1 = cmap1(norm1(img1))
    if norm2 is None:
        norm2 = plt.Normalize()
    cols2 = cmap2(norm2(img2))

    blend_r = np.zeros(cols1.shape) + blend_r
    if isinstance(img1, np.ma.MaskedArray):
        blend_r[np.where(img1.mask)] = 0.

    if isinstance(img2, np.ma.MaskedArray):
        blend_r[np.where(img1.mask)] = 1.

    mixed_cols = blend_r * cols1 + (1.-blend_r) * cols2

    # mixed_cols = np.zeros_like(cols1)
    # for i in xrange(cols1.shape[0]):
    #     for j in xrange(cols1.shape[1]):
    #         br = blend_r[i,j]
    #         mixed_cols[i,j] = (br)*cols1[i,j] + (1.-br)*cols2[i,j]
    #         if 0:
    #             print 'mixing [%d,%d] -> %1.2f * %s + %1.2f * %s = %s' \
    #                 %(i,j, blend_r, str(cols1[i,j]), 1.-blend_r, str(cols2[i,j]),
    #                   str(mixed_cols[i,j]))
    return mixed_cols


def plot_func_slice(func_slice_data, anatomy=None, parcellation=None,
                    parcel_col='white',
                    parcels_line_width=2.5, func_cmap=None,
                    func_norm=None, anat_norm=None, func_mask=None,
                    highlighted_parcels_col=None,
                    highlighted_parcels_line_width=2.5,
                    resolution=None, crop_extension=None, blend=.5):

    import matplotlib.pyplot as plt

    if highlighted_parcels_col is None:
        highlighted_parcels_col = {}

    if func_cmap is None:
        func_cmap = plt.get_cmap('jet')

    if anatomy is not None:
        if func_slice_data.shape != anatomy.shape:
            func_slice_data = rebin(func_slice_data, anatomy.shape)

    if parcellation is not None:
        parcellation = parcellation.astype(int)

        if anatomy is not None:
            if parcellation.shape != anatomy.shape:
                parcellation = rebin(parcellation, anatomy.shape)

        fmask = (parcellation>0).astype(bool)
    else:
        fmask = np.ones_like(func_slice_data).astype(bool)

    if func_mask is not None:
        if anatomy is not None:
            func_mask = rebin(func_mask.astype(bool), anatomy.shape)
         #print 'func_mask:', func_mask.sum()
        fmask = np.bitwise_and(fmask, func_mask)

    if crop_extension is not None:
        from pyhrf.tools import crop_array
        func_slice_data = crop_array(func_slice_data, fmask, crop_extension)
        if anatomy is not None:
            m_tmp = np.where(fmask!=0)
            mask_anat_tmp = np.zeros_like(anatomy)
            e = crop_extension
            start_i = m_tmp[0].min()-e-1
            start_j = m_tmp[1].min()-e-1
            if start_i < 0 or start_j < 0:
                raise Exception('Anatomy is not large enough to extend view.')

            mask_anat_tmp[start_i:m_tmp[0].max()+e,start_j:m_tmp[1].max()+e] = 1
            anatomy = crop_array(anatomy, mask_anat_tmp)
            # print 'func_slice_data:', func_slice_data.shape
            # print 'anatomy:', anatomy.shape
        fmask = crop_array(fmask, fmask, crop_extension)

    func_masked = np.ma.masked_where(fmask==0, func_slice_data)


    anat_cmap = plt.get_cmap('gray')

    plt.figure()
    plt.hold(True)

    if anatomy is not None:
        ax = plt.imshow(mix_cmap(func_masked, func_cmap,
                                 anatomy, anat_cmap,
                                 norm1=func_norm,
                                 norm2=anat_norm, blend_r=blend),
                        interpolation='nearest')
    else:
        ax = plt.imshow(func_masked, cmap=func_cmap,
                        norm=func_norm,
                        interpolation='nearest')
    ax.get_axes().set_axis_off()


    if resolution is not None:
        ratio = resolution[0] / resolution[1]
        print 'Set aspect ratio to:', ratio
        ax.get_axes().set_aspect(ratio)

    if parcellation is not None:
        labs = np.unique(parcellation)

        # nr, nc = parcel_rebin.shape
        # extent = [-0.5, nc-0.5, -0.5, nr-0.5]

        for lab in labs: #[484,427,540]: #
            if lab != 0:
                col = highlighted_parcels_col.get(lab, parcel_col)
                if lab in highlighted_parcels_col:
                    lw = highlighted_parcels_line_width
                else:
                    lw = parcels_line_width
                plt.contour((parcellation==lab).astype(int), 1,
                            colors=[col], linewidths=lw, alpha=.6,)
                #           extent=extent)



def plot_anat_parcel_func_fusion(anat, func, parcel, parcel_col='white',
                                 parcels_line_width=.5, func_cmap=None,
                                 func_norm=None, anat_norm=None, func_mask=None,
                                 highlighted_parcels_col=None,
                                 highlighted_parcels_line_width=1.5):

    import matplotlib.pyplot as plt

    if highlighted_parcels_col is None:
        highlighted_parcels_col = {}

    if func_cmap is None:
        func_cmap = plt.get_cmap('jet')

    parcel = parcel.astype(int)
    parcel_rebin = rebin(parcel, anat.shape)

    #parcel_rebin_ma = np.ma.masked_where(parcel_rebin==0, parcel_rebin)
    func_rebin = rebin(func, anat.shape)
    fmask = (parcel_rebin>0).astype(bool)

    if func_mask is not None:
        func_mask_rebin = rebin(func_mask.astype(bool), anat.shape)
        #print 'func_mask:', func_mask_rebin.sum()
        fmask = np.bitwise_and(fmask, func_mask_rebin)
    func_rebin_ma = np.ma.masked_where(fmask==0, func_rebin)

    anat_cmap = plt.get_cmap('gray')

    plt.figure()
    plt.hold(True)
    if 0:
        #TODO: better color mixing
        ax = plt.imshow(anat, cmap=anat_cmap, norm=anat_norm,
                        interpolation='nearest')

        plt.imshow(func_rebin_ma, interpolation='nearest',
                   alpha=.5, cmap=func_cmap, norm=func_norm, axes=ax)

    if 1:
        ax = plt.imshow(mix_cmap(func_rebin_ma, func_cmap,
                                 anat, anat_cmap,
                                 norm1=func_norm,
                                 norm2=anat_norm, blend_r=.5),
                        interpolation='nearest')
        ax.get_axes().set_axis_off()

    if 0:
        plt.imshow(func_rebin_ma*anat, interpolation='nearest',
                   cmap=func_cmap, norm=func_norm)
    if 0:
        ax = plt.imshow(func_rebin_ma, interpolation='nearest',
                        cmap=func_cmap, norm=func_norm)
        plt.colorbar()
        plt.imshow(anat, cmap=plt.get_cmap('gray'), alpha=.65,
                   interpolation='nearest', axes=ax)

    if 0:
        plt.imshow(mix_cmap(func_rebin_ma, func_cmap, anat, anat_cmap),
                   interpolation='nearest')

    labs = np.unique(parcel)

    # nr, nc = parcel_rebin.shape
    # extent = [-0.5, nc-0.5, -0.5, nr-0.5]

    for lab in labs: #[484,427,540]: #
        if lab != 0:
            col = highlighted_parcels_col.get(lab, parcel_col)
            if lab in highlighted_parcels_col:
                lw = highlighted_parcels_line_width
                alpha = .5
            else:
                lw = parcels_line_width
                alpha = .2
            plt.contour((parcel_rebin==lab).astype(int), 1,
                        colors=[col], linewidths=lw, alpha=alpha,)
    #extent=extent)


def autocrop(img_fn):
    """ Remove extra background within figure (inplace).
    Use ImageMagick (convert) """
    os.system('convert %s -trim %s' %(img_fn,img_fn))

def rotate(img_fn, angle):
    """ Rotate figure (inplace). Use ImageMagick (convert) """
    os.system('convert %s -rotate %f %s' %(img_fn,angle,img_fn))

def flip(img_fn, direction='horizontal'):
    """ Mirror the figure (inplace). Use ImageMagick (convert)
    'horizontal' direction -> use -flop.
    'vertical' direction -> use -flip.
    """
    cmd_arg = {'horizontal':'flop','vertical':'flip'}
    os.system('convert %s -%s %s' %(img_fn,cmd_arg[direction],img_fn))


def plot_palette(cmap, norm=None, fontsize=None):
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    if norm is None:
        norm = plt.Normalize()
    fig = plt.figure()
    ax1 = fig.add_axes([0.05, 0.05, 0.05, .9])
    colbar = mpl.colorbar.ColorbarBase(ax1, cmap=cmap,
                              norm=norm,
                              orientation='vertical')
    if fontsize is not None:
        ticklabels = plt.getp(colbar.ax, 'yticklabels')
        plt.setp(ticklabels, 'color', 'k', fontsize=fontsize)
    return colbar

def set_ticks_fontsize(fontsize, colbar=None):
    """ Change the fontsize of the tick labels for the current figure.
    If colorbar (Colorbar instance) is provided then change the fontsize of its
    tick labels as well.
    """
    import matplotlib.pyplot as plt

    yticklabels = plt.getp(plt.gca(), 'yticklabels')
    plt.setp(yticklabels, 'color', 'k', fontsize=fontsize)
    xticklabels = plt.getp(plt.gca(), 'xticklabels')
    plt.setp(xticklabels, 'color', 'k', fontsize=fontsize)

    if colbar is not None:
        ticklabels = plt.getp(colbar.ax, 'yticklabels')
        plt.setp(ticklabels, 'color', 'k', fontsize=fontsize)


def set_xticklabels(labels, positions=None, rotation=None):
    """ Set ticks labels of the xaxis in the current figure to *labels*
    If positions is provided then enforce tick position.
    """
    import matplotlib.pyplot as plt

    if positions is None:
        positions,_ = plt.xticks()

    if len(positions) != len(labels):
        raise Exception('Wrong number of labels (%d), ticks require %s' \
                        %(len(labels), len(positions)))

    plt.xticks(positions, labels)
    labs = plt.gca().get_xticklabels()
    plt.setp(labs, rotation=rotation)
#     ca = plt.gca().axis
#     plt.setp(labels, 'color', 'k', fontsize=fontsize)



def plot_spm_mip(img_fn, mip_fn):
    pass

import scipy.stats
def plot_gaussian_pdf(bins, m, v, prop=None, plotArgs={}):
    import matplotlib.pyplot as plt

    prop = prop or 1.
    if v > 1e-2:
        pdf = scipy.stats.norm.pdf(bins, m, v)
    else:
        if v < 1e-4: v = 1e-4
        bins = np.arange(bins[0], bins[-1], 0.0001)
        pdf = scipy.stats.norm.pdf(bins, m, v)
    pdf *= prop
    plt.plot(bins, pdf, **plotArgs)
    return pdf

# lw = 1.75
# fontsize=18

def plot_gaussian_mixture(values, props=None, color='k', lw=1.75):
    """
    axes of values : (component,class)
    """
    import matplotlib.pyplot as plt

    if props is None:
        props = np.array([.5, .5])
    xMin = (values[0,:] - 5*values[1,:]**.5).min()
    xMax = (values[0,:] + 5*values[1,:]**.5).max()
    nbins = 100.
    bins = np.arange(xMin, xMax, (xMax-xMin)/nbins)
    plotArgs = {'linestyle':'--',
                'linewidth':lw,
                'color':color}
    pdf0 = plot_gaussian_pdf(bins, values[0,0], values[1,0],
                             props[0], plotArgs)
    plotArgs = {'linestyle':'-',
                'linewidth':lw,
                'color':color}
    pdf1 = plot_gaussian_pdf(bins, values[0,1], values[1,1],
                             props[1], plotArgs)

    plotArgs = {'linestyle':'-.',
                'linewidth':lw,
                'color':color}
    plt.plot(bins, pdf0 + pdf1, **plotArgs)

    if plt.ylim()[1] > 2.0:
        plt.ylim(0, 2.)
    #nyticks =  len(plt.yticks())
    #plt.yticks(np.arange(nyticks), ['']*nyticks)


def plot_cub_as_curve(c, colors=None, plot_kwargs=None, legend_prefix='', show_axis_labels=True, show_legend=False,
                      axes=None, axis_label_fontsize=12):
    """Plot a cuboid (ndims <= 2) as curve(s).

    - If the input is 1D: one single curve.
    - If the input is 2D:
        * multiple curves are plotted: one for each domain value on the 1st axis.
        * legends are shown to display which domain value is associated
          to which curve.

    Parameters
    ----------
    colors : dict <domain value>: <matplotlib color>
        associate domain values of the 1st axis to color curves
    plot_kwargs : dict <arg name>:<arg value>
        dictionary of named argument passed to the plot function
    legend_prefix : str
        prefix to prepend to legend labels.

    Returns
    -------
    None
    """
    import matplotlib.pyplot as plt

    def protect_latex_str(s):
        return s.replace('_','\_')
        
    axes = axes or plt.gca()
    colors = colors or {}
    plot_kwargs = plot_kwargs or {}
    if c.get_ndims() == 1:
        dom = c.axes_domains[c.axes_names[0]]
        if np.issubsctype(dom.dtype, str):
            dom = np.arange(len(dom))
        axes.plot(dom, c.data, **plot_kwargs)
        if np.issubsctype(c.axes_domains[c.axes_names[0]], str):
            set_int_tick_labels(axes.xaxis, c.axes_domains[c.axes_names[0]],
                                rotation=30)
    elif c.get_ndims() == 2:
        for val, sub_c in c.split(c.axes_names[0]).iteritems():
            pkwargs = plot_kwargs.copy()
            col = colors.get(val, None)
            if col is not None:
                pkwargs['color'] = col
            pkwargs['label'] = protect_latex_str(legend_prefix + \
                                                 c.axes_names[0] + \
                                                 '=' + str(val))
            plot_cub_as_curve(sub_c, plot_kwargs=pkwargs, axes=axes,
                              show_axis_labels=False)
        if show_legend:
            axes.legend()

    else:
        raise Exception('xndarray has too many dims (%d), expected at most 2' \
                        %c.get_ndims())

    if show_axis_labels:
        if c.get_ndims() == 1:
            axes.set_xlabel(protect_latex_str(c.axes_names[0]), 
                            fontsize=axis_label_fontsize)
        else:
            axes.set_xlabel(protect_latex_str(c.axes_names[1]),
                            fontsize=axis_label_fontsize)
        axes.set_ylabel(protect_latex_str(c.value_label),
                        fontsize=axis_label_fontsize)

def set_int_tick_labels(axis, labels, fontsize=None, rotation=None):
    """
    Redefine labels of visible ticks at integer positions for the given axis.
    """
    # get the tick positions:
    tickPos = axis.get_ticklocs()#.astype(int)
    dvMax = len(labels)
    tLabels = []
    #if debug: print '%%%% tickPos :', tickPos
    for tp in tickPos:
        if tp < 0. or int(tp) != tp or tp >= dvMax:
            tLabels.append('')
        else:
            tLabels.append(labels[int(tp)])
    #if debug: print '%%%% Setting labels:', tLabels
    axis.set_ticklabels(tLabels)
    for label in axis.get_ticklabels():
        if fontsize is not None:
            label.set_fontsize(fontsize)
        if rotation is not None:
            label.set_rotation(rotation)


def plot_cub_as_image(c, cmap=None, norm=None, show_axes=True,
                      show_axis_labels=True, show_tick_labels=True,
                      show_colorbar=False, axes=None):
    import matplotlib.pyplot as plt

    axes = axes or plt.gca()
    data = c.data
    if data.ndim == 1:
        data = data[:,np.newaxis]

    ms = axes.matshow(data, cmap=cmap, norm=norm)#, origin='lower')
    if show_tick_labels:
        set_int_tick_labels(axes.yaxis, c.axes_domains[c.axes_names[0]])
        if len(c.axes_domains) > 1:
            set_int_tick_labels(axes.xaxis, c.axes_domains[c.axes_names[1]])
    else:
        set_int_tick_labels(axes.yaxis,
                            [''] * len(c.axes_domains[c.axes_names[0]]))
        if len(c.axes_domains) > 1:
            set_int_tick_labels(axes.xaxis,
                                [''] * len(c.axes_domains[c.axes_names[1]]))

    if show_axis_labels:
        axes.set_ylabel(c.axes_names[0])
        if len(c.axes_domains) > 1:
            axes.set_xlabel(c.axes_names[1])
    else:
        axes.set_ylabel('')
        axes.set_xlabel('')

    if not show_axes:
        axes.set_axis_off()

    if show_colorbar:
        axes.figure.colorbar(ms)



# -*- coding: utf-8 -*-
"""PyHRF is a set of tools for within-subject fMRI data analysis,
which focuses on the characterization of the hemodynamics.

Within the chain of fMRI data processing, these tools provide alternatives to
the classical within-subject GLM estimation step. The inputs are preprocessed
within-subject data and the outputs are statistical maps and/or fitted HRFs.

The package is mainly written in Python and provides the implementation of
the two following methods:

- The joint-detection estimation (JDE) approach, which divides the brain
  into functionally homogeneous regions and provides one HRF estimate per
  region as well as response levels specific to each voxel and each experimental
  condition. This method embeds a temporal regularization on the estimated HRFs
  and an adaptive spatial regularization on the response levels.

- The Regularized Finite Impulse Response (RFIR) approach, which provides
  HRF estimates for each voxel and experimental conditions. This method embeds
  a temporal regularization on the HRF shapes, but proceeds independently across
  voxels (no spatial model).

Check the PyHRF `website <http://www.pyhrf.org>`_ for details.
"""

import os
import logging
import warnings
import sys

import numpy

import pyhrf._verbose

from pyhrf.version import __version__
from pyhrf.configuration import cfg, useModesStr
from pyhrf.configuration import DEVEL, ENDUSER


# Resolve current version
_VERSION_STRING = 'pyhrf ' + __version__
use_modes = useModesStr.keys()
if cfg['global']['use_mode'] not in use_modes:
    raise Exception('Wrong use mode "%s" in configuration, '
                    'available choices: %s' % (cfg['global']['use_mode'],
                                               ', '.join(use_modes)))

__usemode__ = useModesStr[cfg['global']['use_mode']]

verbose_levels = {logging.getLevelName(value): value
                  for value in dir(logging)
                  if value.isupper() and
                  isinstance(logging.getLevelName(value), int)}

numpy.seterr(cfg['global']['numpy_floating_point_error'])

for pname, pval in pyhrf.configuration.cfg['global'].iteritems():
    setattr(pyhrf.configuration, pname, pval)

# logging configuration
# warnings.filterwarnings("always", message=".*verbos.*",
#                         category=DeprecationWarning)
logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger(__name__)
verb = cfg['log']['level']
try:
    logger.setLevel(verb.upper())
except ValueError, e:
    verb = logging.WARNING
    logger.setLevel(verb)
    logger.error(e.message + '. Check your configuration file.'
                 ' Falling back to default WARNING level')
if cfg['log']['log_to_file']:
    file_handler = logging.FileHandler(os.path.join(cfg['global']['tmp_path'],
                                                    'pyhrf.log'))
    formatter = logging.Formatter(
        '%(asctime)s :: %(levelname)s :: %(name)s :: %(message)s')
    file_handler.setLevel(verb)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)


class Verbose(pyhrf._verbose.Verbose):
    """This is a dummy class implementing the original Verbose class.

    This is only to be able to raise a warning when one uses this old
    implementation.
    """

    old_to_new_log_dict = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.INFO,
        3: logging.INFO,
        4: logging.INFO,
        5: logging.DEBUG,
        6: logging.DEBUG,
    }

    def __call__(self, verbosity, msg, new_line=True):
        warnings.warn('The homemade verbose module is deprecated,'
                      ' use logging instead', DeprecationWarning)
        logger.log(Verbose.old_to_new_log_dict[verbosity], msg)

verbose = Verbose(verbosity=verb)

# FIXME: The following are two things we want to avoid:
# import not at the beginning and wildcard import
# and furthermore it erase the logger variable in the namespace
tmplogger = logger
from pyhrf.core import *
logger = tmplogger



# -*- coding: utf-8 -*-

import unittest
import shutil
import tempfile
import logging

import numpy as np

try:
    from sklearn.externals.joblib import Memory
except ImportError:
    def Memory(*args, **kwargs):
        pass

import pyhrf

from pyhrf.core import FmriData
from pyhrf.ui.treatment import FMRITreatment
from pyhrf.ui.vb_jde_analyser_asl_fast import JDEVEMAnalyser


logger = logging.getLogger(__name__)


class ASLTest(unittest.TestCase):

    def setUp(self):
        cache_dir = tempfile.mkdtemp(prefix='pyhrf_validate',
                                     dir=pyhrf.cfg['global']['tmp_path'])
        mem = Memory(cache_dir)
        np.random.seed(8652761)

        self.tmp_dir = pyhrf.get_tmp_path()
        self.clean_tmp = True

    def tearDown(self):
        if self.clean_tmp:
            logger.info('Remove tmp dir %s', self.tmp_dir)
            shutil.rmtree(self.tmp_dir)
            del mem
            shutil.rmtree(cache_dir)
        else:
            logger.info('Keep tmp dir %s', self.tmp_dir)

    def test_prf(self):
        """ Validate estimation of PRF """
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir)
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['prf'], fdata, simu, nItMax=20,
                                       estimateG=True)

    def test_brf(self):
        """ Validate estimation of BRF at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir)
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['brf'], fdata, simu, nItMax=100,
                                       estimateH=True)

    def test_brls(self):
        """ Validate estimation of BRLs at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir)
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['bold_response_levels'], fdata, simu,
                                       nItMax=100, estimateA=True)

    def test_prls(self):
        """ Validate estimation of PRLs at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir)
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['perf_response_levels'], fdata, simu,
                                       nItMax=100, estimateC=True)

    def test_labels(self):
        """ Validate estimation of labels at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir)
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['labels'], fdata, simu, nItMax=100,
                                       estimateZ=True)
        print 'pyhrf_view %s/*label*nii' % self.tmp_dir

    def test_noise_var(self):
        """ Validate estimation of noise variances at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir)
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['noise_var'], fdata, simu, nItMax=100,
                                       estimateNoise=True)
        print 'pyhrf_view %s/*noise*nii' % self.tmp_dir

    def test_la(self):
        """ Validate estimation of drift at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['drift_perf_baseline'], fdata, simu,
                                       nItMax=100, estimateLA=True)
        print 'pyhrf_view %s/*drift*nii' % self.tmp_dir

    def test_mp(self):
        """ Validate estimation of drift at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['mixture_params'], fdata, simu,
                                       nItMax=100, estimateMP=True)
        print 'pyhrf_view %s/*drift*nii' % self.tmp_dir

    def test_beta(self):
        """ Validate estimation of drift at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['beta'], fdata, simu,
                                       nItMax=100, estimateBeta=True)
        print 'pyhrf_view %s/*beta*nii' % self.tmp_dir

    def test_sigmaH(self):
        """ Validate estimation of drift at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['sigma_H'], fdata, simu,
                                       nItMax=100, estimateSigmaH=True)
        print 'pyhrf_view %s/*mixt_params*b*nii' % self.tmp_dir

    def test_sigmaG(self):
        """ Validate estimation of drift at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        self._test_specific_parameters(['sigma_G'], fdata, simu,
                                       nItMax=100, estimateSigmaG=True)
        print 'pyhrf_view %s/*mixt_params*perf*nii' % self.tmp_dir

    def test_bold(self):
        """ Validate estimation of bold component at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        np.random.seed(25430)
        v = ['bold_response_levels', 'brf']

        mem.cache(self._test_specific_parameters)(v, fdata, simu,
                                                  nItMax=100,
                                                  estimateH=True,
                                                  estimateA=True,
                                                  estimateSigmaH=True)
        print 'pyhrf_view %s/*nii' % self.tmp_dir

    def test_perfusion(self):
        """ Validate estimation of perfusion component at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        np.random.seed(25430)
        v = ['perf_response_levels', 'prf']

        mem.cache(self._test_specific_parameters)(v, fdata, simu,
                                                  nItMax=100,
                                                  estimateG=True,
                                                  estimateC=True,
                                                  estimateSigmaG=True)
        print 'pyhrf_view %s/*nii' % self.tmp_dir

    def test_E_step(self):
        """ Validate estimation of perfusion component at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        np.random.seed(25430)
        v = ['perf_response_levels', 'prf']

        mem.cache(self._test_specific_parameters)(v, fdata, simu,
                                                  nItMax=100,
                                                  estimateH=True,
                                                  estimateA=True,
                                                  estimateSigmaH=True,
                                                  estimateG=True,
                                                  estimateC=True,
                                                  estimateSigmaG=True,
                                                  estimateZ=True)
        print 'pyhrf_view %s/*nii' % self.tmp_dir

    def test_all(self):
        """ Validate estimation of full ASL model at high SNR"""
        # pyhrf.verbose.set_verbosity(2)
        pyhrf.logger.setLevel(logging.INFO)
        from pyhrf.jde.asl import simulate_asl
        simu = simulate_asl(self.tmp_dir, spatial_size='normal')
        fdata = FmriData.from_simulation_dict(simu)
        np.random.seed(25430)
        v = ['bold_response_levels', 'perf_response_levels',
             'brf', 'brf_var', 'prf', 'labels', 'bold_mixt_params',
             'perf_mixt_params', 'drift_perf_baseline']

        self._test_specific_parameters(v, fdata, simu,
                                       estimateSigmaH=False, nItMax=100,
                                       nItMin=10, estimateBeta=True,
                                       estimateSigmaG=True, PLOT=False,
                                       constrained=True, fast=False,
                                       estimateH=True, estimateG=True,
                                       estimateA=True, estimateC=True,
                                       estimateZ=True, estimateLA=True,
                                       estimateMP=True)
        print 'pyhrf_view %s/*nii' % self.tmp_dir

    def _test_specific_parameters(self, parameter_name, fdata, simu,
                                  beta=.8, dt=.5, nItMax=100, nItMin=10,
                                  hrfDuration=25., estimateSigmaH=False,
                                  estimateBeta=False, estimateSigmaG=False,
                                  PLOT=False, constrained=True, fast=False,
                                  estimateH=False, estimateG=False,
                                  estimateA=False, estimateC=False,
                                  estimateZ=False, estimateLA=False,
                                  estimateNoise=False, estimateMP=True):
        """
        Test specific samplers.
        """
        logger.info('_test_specific_parameters %s', str(parameter_name))
        output_dir = self.tmp_dir
        # JDE analysis
        jde_vem_analyser = JDEVEMAnalyser(beta=beta, dt=dt,
                                          hrfDuration=hrfDuration,
                                          estimateSigmaH=estimateSigmaH,
                                          nItMax=nItMax, nItMin=nItMin,
                                          estimateBeta=estimateBeta,
                                          estimateSigmaG=estimateSigmaG,
                                          PLOT=PLOT,
                                          constrained=constrained, fast=fast,
                                          fmri_data=fdata,
                                          simulation=simu,
                                          estimateH=estimateH,
                                          estimateG=estimateG,
                                          estimateA=estimateA,
                                          estimateC=estimateC,
                                          estimateLabels=estimateZ,
                                          estimateLA=estimateLA,
                                          estimateMixtParam=estimateMP,
                                          estimateNoise=estimateNoise)
        tjde_vem = FMRITreatment(fmri_data=fdata, analyser=jde_vem_analyser,
                                 output_dir=output_dir)
        outputs = tjde_vem.run()
        print 'out_dir:', output_dir
        return outputs



# -*- coding: utf-8 -*-


from bigaussian import NRLSampler
from bigaussian import BiGaussMixtureParamsSampler
from bigaussian import MixtureWeightsSampler
from bigaussian_drift import NRL_Drift_Sampler

from habituation import NRLwithHabSampler

from trigaussian import GGGNRLSampler, TriGaussMixtureParamsSampler

from gammagaussian import InhomogeneousNRLSampler, GamGaussMixtureParamsSampler

from ar import NRLARSampler



# -*- coding: utf-8 -*-

import copy as copyModule
import hashlib
import logging

import numpy as NP

from numpy import cos, sin

from pyhrf.tools import cartesian
from pyhrf.ndarray import *


logger = logging.getLogger(__name__)


def hashMask(m):
    return hashlib.sha512(m.astype(int).tostring()).hexdigest()


def getRotationMatrix(axis, angle):
    """
    Compute the  3x3 matrix for the 3D rotation defined by 'angle' and the
    direction 'axis'.
    """
    ca = cos(angle)
    sa = sin(angle)
    mc = 1 - ca

    v = NP.array(axis)
    n = NP.sqrt(NP.sum(v ** 2))
    # Ensure that axis is a unit vector :
    x = v[0] / n
    y = v[1] / n
    z = v[2] / n
    return NP.array([[ca + mc * x * x, mc * x * y - sa * z, mc * x * z + sa * y],
                     [mc * y * x + sa * z, ca + mc * y * y,
                         mc * y * z - sa * x],
                     [mc * z * x - sa * y, mc * z * y + sa * x, ca + mc * z * z]])

####################
# Spatial Mappings #
####################

# TODO: maybe make it recursive to manage different flatten level?
# -> better to use chain iterator


def flattenElements(l):
    dest = []
    for le in l:
        dle = []
        for e in le:
            if np.iterable(e):
                dle.extend(e)
            else:
                dle.append(e)
        dest.append(dle)
    return dest


class Mapper1D:
    """
    Handles a mapping between a nD coordinate space (expanded) and a 1D
    coordinate space (flatten). Can be applied to numpy.ndarray objects
    """

    def __init__(self, mapping, expandedShape):
        """
        Initialise a Mapper1D:
        'mapping' is a list of coordinates in the expanded space.
        'expandedShape' is the shape of the expanded space.
        """
        assert len(mapping[0]) == len(expandedShape)
        self.mapping = mapping
        self.expandedShape = expandedShape

    def flattenArray(self, array, firstMappedAxis=0):
        """
        Reduce dimensions of 'array'. 'firstMappedAxis' is index of the axis to
        be reduced (other mapped axes are assumed to follow this one).
        """
        nd = len(self.expandedShape)
        ia = firstMappedAxis
        assert len(self.expandedShape) <= array.ndim
        assert self.expandedShape == array.shape[ia:ia + nd]
        ranges = [range(d) for d in array.shape]

        ranges = ranges[:ia] + [self.mapping] + ranges[ia + nd:]
        coords = cartesian(*ranges)
        coords = flattenElements(coords)
        # TODO: remove explicit loop!
        mask = [NP.array(a) for a in NP.array(coords).transpose()]
        destshape = array.shape[:ia] + \
            (len(self.mapping),) + array.shape[ia + nd:]
        destData = NP.empty(destshape, dtype=array.dtype)
        destData.flat = array[mask]
        return destData

    def createExpandedArray(self, flatShape, type, mappedAxis=0, fillValue=0):
        ia = mappedAxis
        destshape = flatShape[:ia] + self.expandedShape + flatShape[ia + 1:]
        logger.debug('dest shape: %s', str(destshape))
        return NP.zeros(destshape, dtype=type) + fillValue

    def expandArray(self, a, mappedAxis=0, dest=None, fillValue=0):
        """
        Expand dimensions of 'a' following predefined mapping. 'mappedAxis'
        is the axis index to be expanded in 'a'.
        If dest is not None the map values from 'a' to dest.
        If dest is None then return a new array and fill positions not involved
        in mapping with 'fillValue'.
        """
        ia = mappedAxis
        m = self.mapping.transpose()
        lmask = [':' for i in xrange(a.ndim + len(m) - 1)]
        for i in xrange(len(m)):
            lmask[ia + i] = 'm[%d]' % i

        if dest is None:
            dest = self.createExpandedArray(a.shape, a.dtype, mappedAxis,
                                            fillValue)

        if a.shape[mappedAxis] == 0 and len(self.mapping) > 1:
            # assume diplucation
            a = NP.repeat(a, len(self.mapping[0]), mappedAxis)

        sexec = "dest[%s] = a" % ','.join(lmask)
        exec(sexec)

        return dest


class xndarrayMapper1D(Mapper1D):

    def __init__(self, mapping, expandedShape, expandedAxesNames, flatAxisName):
        Mapper1D.__init__(self, mapping, expandedShape)
        self.expandedAxesNames = expandedAxesNames
        self.flatAxisName = flatAxisName

    def isExpendable(self, c):
        return self.flatAxisName in c.getAxesNames()

    def isFlattenable(self, c):
        return set(self.expandedAxesNames).issubset(c.getAxesNames())

    def flattenxndarray(self, c):

        expDim = len(self.expandedAxesNames)
        fia = c.getAxisId(self.expandedAxesNames[0])
        can = c.getAxesNames()
        assert (can[fia:fia + expDim] == self.expandedAxesNames)

        destAN = can[:fia] + [self.flatAxisName] + can[fia + expDim:]
        destAD = c.getAxesDomainsDict()
        for a in self.expandedAxesNames:
            destAD.pop(a, None)
        flatData = self.flattenArray(c.data, fia)
        flatErrors = None
        if c.errors is not None:
            flatErrors = self.flattenArray(c.errors, fia)
        return xndarray(flatData, flatErrors, axes_names=destAN,
                        axes_domains=destAD, value_label=c.value_label)

    def expandxndarray(self, c, dest=None, fillValue=0):
        logger.info('expandxndarray ...')
        ia = c.getAxisId(self.flatAxisName)
        can = c.getAxesNames()
        logger.info('can : %s', str(can))
        nd = len(self.expandedAxesNames)
        destAN = can[:ia] + self.expandedAxesNames + can[ia + 1:]
        logger.info('destAN : %s', str(destAN))
        destAD = c.getAxesDomainsDict()
        destAD.pop(self.flatAxisName, None)
        if dest is None:
            destData = self.createExpandedArray(c.data.shape, c.data.dtype,
                                                ia, fillValue)
            destErrors = None
            if c.errors is not None:
                destErrors = self.createExpandedArray(c.data.shape,
                                                      c.errors.dtype, ia,
                                                      fillValue)
            dest = xndarray(destData, destErrors, axes_names=destAN,
                            axes_domains=destAD, value_label=c.value_label)

        logger.info('src shape : %s, -> des shape: %s', str(c.data.shape),
                    str(dest.data.shape))
        Mapper1D.expandArray(self, c.data, ia, dest.data, fillValue)
        if c.errors is not None:
            Mapper1D.expandArray(self, c.errors, ia, dest.errors, fillValue)

        return dest


def lattice_indexes(mask):
    mpos = NP.where(mask)
    indexes = NP.zeros_like(mask).astype(int) - 1
    indexes[mpos] = range(len(mpos[0]))
    return indexes


class NeighbourhoodSystem:

    kerMask3D_6n = NP.array([[1, 0, 0], [0, 1, 0], [0, 0, 1],
                             [-1, 0, 0], [0, -1, 0], [0, 0, -1]], dtype=int)

    kerMask2D_4n = NP.array([[-1, 0], [1, 0], [0, 1], [0, -1]], dtype=int)

    def __init__(self, neighboursSets):
        self.neighboursSets = neighboursSets
        self.maxNeighbours = None
        self.neighboursArray = None

    def getNeighboursLists(self):
        return NP.asarray([list(ns) for ns in self.neighboursSets], dtype=object)

    def getNeighboursSets(self):
        return self.neighboursSets

    def getMaxNeighbours(self):
        if self.maxNeighbours == None:
            m = 0
            # for nrl in self.neighboursSets.itervalues():
            for nl in self.neighboursSets:
                m = len(nl) if len(nl) > m else m
            self.maxNeighbours = m
        return self.maxNeighbours

    def getMaxIndex(self):
        if self.maxIndex == None:
            self.maxIndex = len(self.neighboursSets) - 1
        return self.maxIndex

    def getNeighboursArrays(self):
        if self.neighboursArray == None:
            mn = self.getMaxNeighbours()
            mi = self.getMaxIndex()
            self.neighboursArray = NP.zeros((mi + 1, mn), dtype=int) - 1
            idx = 0
            for nl in self.neighboursSets:
                self.neighboursArray[idx, :len(nl)] = nl
                idx += 1
        return self.neighboursArray

    def getNeighbours(self, nodeId):
        return self.neighboursSets[nodeId]

    def sub(self, nodeIds):

        # Filter neighbourslists to only keep given nodes:
        sNodeIds = set(nodeIds)
        newNL = NP.array([sNodeIds.intersection(nl)
                          for nl in self.neighboursSets[nodeIds]])
        return NeighbourhoodSystem(newNL)

    @staticmethod
    def fromMesh(polygonList):
        nls = graphFromMesh(polygonList)
        ns = NP.empty(max(nl.iterkeys()), dtype=object)
        for idx, nl in nls:
            ns[idx] = nl
        return NeighbourhoodSystem(ns)

    @staticmethod
    def fromLattice(latticeIndexes, kerMask=None, depth=1, torusFlag=False):
        """
        Creates a NeighbourhoodSystem instance from a n-dimensional lattice
        """
        mpos = NP.where(latticeIndexes != -1)
        positions = NP.vstack(mpos).transpose()

        if kerMask == None:
            # Full neighbourhood, ie 8 in 2D, 26 in 3D ...
            ndim = latticeIndexes.ndim
            kerMask = NP.array(
                list(cartesian(*[[0, -1, 1]] * ndim))[1:], dtype=int)

        mapper = LatticeIndexMapper(latticeIndexes, torusFlag)
        nbPositions = latticeIndexes.size

        # Build lists of closest neighbours:
        closestNeighbours = NP.empty(nbPositions, dtype=object)
        for idx, pos in enumerate(positions):
            ni = map(mapper.indexFromCoord, kerMask + pos)
            closestNeighbours[idx] = set(filter(lambda x: x is not None, ni))

        if depth == 1:
            return NeighbourhoodSystem(closestNeighbours)

        neighboursSets = copyModule.deepcopy(closestNeighbours)

        for idx in xrange(nbPositions):
            neighboursToTreat = closestNeighbours[idx]
            visited = set([idx])
            for d in xrange(depth - 1):
                newNeighboursToTreat = set()
                # In the worst case, neighbours to treat are on the perimeter
                # of the lattice, say O(sum(lattice shape))
                # OR
                # If the lattice is kind of flat, the worst case is
                # O(size of kerMask)
                # So complexity is :
                # O(max( sum(lattice shape), kerMask size)) * O(kerMask size)
                for n in neighboursToTreat:
                    # O(kerMask size)
                    newNeighboursToTreat.update(closestNeighbours[n])
                # Remember what we just saw:
                visited.update(neighboursToTreat)
                # Update neighbours of current position and remove those
                # already seen:
                newNeighboursToTreat.difference_update(visited)
                neighboursSets[idx].update(newNeighboursToTreat)
                # Define new starting positions for next loop:
                neighboursToTreat = newNeighboursToTreat

        return NeighbourhoodSystem(neighboursSets)


class SpatialMapping2:

    def __init__(self, positions, ns, parentMapping=None, parentIndex=None):
        self.ns = ns
        self.positions = positions
        self.parent = parentMapping
        self.parentIndex = parentIndex

    def getPositions(self):
        return self.positions

    def sub(self, nodeIds):
        subPos = self.positions[nodeIds]
        return SpatialMapping2(subPos, self.ns.sub(nodeIds), self)

    @staticmethod
    def fromMesh(triangles, positions):
        return GraphMapping(NeighbourhoodSystem.fromMesh(triangles), positions)

    @staticmethod
    def fromLattice(lattice, kerMask=None, nsDepth=1, torusFlag=False):
        mpos = NP.where(lattice)
        positions = NP.vstack(mpos).transpose()
        ns = NeighbourhoodSystem.fromLattice(lattice, kerMask=kerMask,
                                             depth=nsDepth, torusFlag=torusFlag)
        return SpatialMapping2(ns, positions)


class RegularLatticeMapping2(SpatialMapping2):

    def __init__(self, maskLattice, kerMask=None, nsDepth=1, parentMapping=None,
                 torusFlag=False):
        self.nbDims = maskLattice.ndim
        self.targetShape = maskLattice.shape
        indexLattice = NP.zeros_like(maskLattice)
        m = NP.where(maskLattice)
        ids = range(len(m[0]))
        indexLattice[m] = ids
        positions = NP.vstack(m).transpose()
        ns = NeighbourhoodSystem.fromLattice(indexLattice, kerMask=kerMask,
                                             depth=nsDepth, torusFlag=torusFlag)
        SpatialMapping2.__init__(self, positions, ns)

    def flattenData(self, data, firstMappedAxis=0):
        # Note: very similar to pyhrf.ndarray.buildMappingMask ...
        # TODO: factorize axis expansion stuffs !!!
        ranges = [range(d) for d in data.shape]
        ia = firstMappedAxis
        nd = self.nbDims
        ranges = ranges[:ia] + [self.positions] + ranges[ia + nd:]
        coords = cartesian(*ranges)
        coords = flattenElements(coords)
        # TODO: remove explicit loop!
        mask = [NP.array(a) for a in NP.array(coords).transpose()]
        destshape = data.shape[:ia] + \
            (len(self.positions),) + data.shape[ia + nd:]
        destData = NP.empty(destshape, dtype=data.dtype)
        destData.flat = data[mask]
        return destData

    def mapData(self, data, mappedAxis=0, fillValue=0):
        ia = mappedAxis
        nd = self.nbDims
        ranges = [range(d) for d in data.shape]
        ranges = ranges[:ia] + [self.positions] + ranges[ia + 1:]
        coords = cartesian(*ranges)
        coords = flattenElements(coords)
        # TODO: remove explicit loop!
        mask = [NP.array(a) for a in NP.array(coords).transpose()]
        destshape = data.shape[:ia] + self.targetShape + data.shape[ia + 1:]
        destData = NP.zeros(destshape, dtype=data.dtype) + fillValue
        destData[mask] = data.flat
        return destData


class SpatialMapping:
    """
    Interface specification for the handling of a mapping between integer indexes
    and positions in a 3D space.
    """

    def getNbVoxels(self):
        """
        Return the total number of mapped position
        """
        raise NotImplementedError

    def getNdArrayMask(self):
        """
        Return the set of mapped 3D coordinates in a tuple usable as a mask for
        numpy.ndarray
        """
        raise NotImplementedError

    def getRoiMask(self):
        """
        Return a binary or n-ary mask which has the shape of the target data
        """
        raise NotImplementedError

    def getMapping(self):
        """
        Return a mapping object (list or dict) which maps an integer index to
        its 3D coordinates.
        """
        raise NotImplementedError

    def getIndex(coord):
        """
        Return index mapped with 'coord'
        """
        raise NotImplementedError

    def getCoord(index):
        """
        Return coord mapped with 'index'
        """
        raise NotImplementedError

    def getNeighboursIndexes(self, idVoxel):
        """
        @param idVoxel: index of the voxel
        @return: the list of integer indexes corresponding to the neighbours
        of the specified voxel.
        """
        raise NotImplementedError

    def getNeighboursIndexLists(self):
        """
        Get lists of neighbours for all positions
        @param idVoxel: index of the voxel
        @return: a mapping object (list or dict) which maps each integer index
        to a list of integer indexes (the neighbours).
        """
        raise NotImplementedError

    def getNeighboursCoords(self, idVoxel):
        """
        @param idVoxel: index of the voxel
        @return: the list of 3D coordinates corresponding to the neighbours
        of the specified voxel.
        """
        raise NotImplementedError

    def getNeighboursCoordLists(self):
        """
        Get lists of neighbours for all positions
        @param idVoxel: index of the voxel
        @return: a mapping object (list or dict) which maps each integer index
        to a list of 3D coordinates (the neighbours).
        """
        raise NotImplementedError


class UnboundSpatialMapping(SpatialMapping):
    """
    Convinient class to provide an implementation of SpatialMapping when there
    is no mapping.
    """

    def __init__(self, nbVoxels=100):
        self.nbVoxels = nbVoxels
        self.neighboursIndexLists = [[] for i in xrange(nbVoxels)]
        self.neighboursCoordLists = [[] for i in xrange(nbVoxels)]

    def __getinitkwargs__(self):
        return getInit(self, self.variables)

    def getNbVoxels(self):
        return self.nbVoxels

    def getMapping(self):
        return None  # TODO check usability

    def getNdArrayMask(self):
        return None  # TODO check usability

    def getRoiMask(self):
        """
        Return a binary or n-ary 3D mask which has the shape of the target data
        """
        return NP.ones((1, 1, self.nbVoxels), dtype=int)

    def getIndex(self, coord):
        return None

    def getCoord(self, index):
        return None

    def getNeighboursIndexes(self, idVoxel):
        return []

    def getNeighboursCoords(self, idVoxel):
        return []

    def getNeighboursIndexLists(self):
        return self.neighboursIndexLists

    def getNeighboursCoordLists(self):
        return self.neighboursCoordLists

    @staticmethod
    def createFromGUI(GUIobject):
        """
        Creates the actual object based on the parameters
        """
        params = {}
        params[UnboundSpatialMapping.P_NBVOX] = getattr(
            GUIobject, UnboundSpatialMapping.P_NBVOX)
        return UnboundSpatialMapping(**params)


class RegularLatticeMapping(SpatialMapping):
    """
    Define a SpatialMapping on a 3D regular lattice.
    """

    order1Mask = NP.array([[1, 0, 0], [0, 1, 0], [0, 0, 1],
                           [-1, 0, 0], [0, -1, 0], [0, 0, -1]])
    nbNeighboursOrder1 = 6

    order2Mask = NP.array(
        [c for c in cartesian([0, -1, 1], [0, -1, 1], [0, -1, 1])][1:])
    nbNeighboursOrder2 = 26

# TODO AL: handle mapping better
# TODO AL: hide advanced options
# )

    # TODO : add handling of periodic coordinate systems like a torus ...
    # TODO : check diff between order and depth ...
    # , **kwargs):
    def __init__(self, shape=None, mapping=None, order=1, depth=1):
        """
        Construct a spatial configuration corresponding to a regular 3D lattice.
        'shape' is a 3-item sequence defining the X, Y and Z sizes.
        Note on aims dimension conventions :
        (axial, coronal, sagittal)
        'mapping' is a list of integer coordinates defining which positions
                are valid in the regular lattice.
        'order' is the order of the neighbour system in 3D. Examples :
                - if order=1 => 6 neighbour system (with depth=1)
                - if order=2 => 26 neighbour system (with depth=1)
        'depth' is the depth of the neighbour system.

        If 'mapping'=None, the default lattice is full and fill order is X->Y->Z.
        """

        if shape != None:
            assert len(shape) == 3
        else:
            shape = (1, 10, 10)

        assert len(shape) == 3
        assert (order == 1 or order == 2)

        self.shape = tuple(shape)
        self.depth = depth

        if order == 1:
            self.neighbourMask = self.order1Mask
            self.nbNeighboursMax = self.nbNeighboursOrder1
        else:
            self.neighbourMask = self.order2Mask
            self.nbNeighboursMax = self.nbNeighboursOrder2

        if mapping == None:
            # If no mapping is defined then the lattice is "full" :
            self.nbVoxels = shape[0] * shape[1] * shape[2]
            # Default fill order of the lattice is : X->Y->Z :
            lxly = shape[0] * shape[1]
            lx = shape[0]
            ly = shape[1]
            self.mapping = NP.array([[(i % (lxly)) / ly, (i % (lxly)) % ly, i / (lxly)]  # creates a mapping by affecting a voxel position to each number corresponding to a voxel
                                     for i in xrange(self.nbVoxels)])
        else:
            # Else valid positions are defined with 'mask' :
            self.mapping = mapping
            self.nbVoxels = len(mapping)

        mask = NP.array(self.mapping)
        # Center working origin on the "upper left corner" :
        self.correctedMask = mask - mask.min(0)
        # TODO : assert that corrected mask fits in volume with given shape

        self.ndarrayMask = (NP.array([c[0] for c in self.mapping]),
                            NP.array([c[1] for c in self.mapping]),
                            NP.array([c[2] for c in self.mapping]))
        self.ndarrayCorMask = (NP.array([c[0] for c in self.correctedMask]),
                               NP.array([c[1] for c in self.correctedMask]),
                               NP.array([c[2] for c in self.correctedMask]))

        # Build the lattice, positions with '-1' are not valid :
        self.lattice = NP.zeros((self.shape), dtype=int) - 1

        # Fill lattice with integer indexes corresponding to valid positions :
        for iv in xrange(self.nbVoxels):
            x = self.correctedMask[iv, 0]
            y = self.correctedMask[iv, 1]
            z = self.correctedMask[iv, 2]
            # Fill lattice with index corresponding to a valid position :
            self.lattice[x, y, z] = iv

        self.neighboursCoordLists = None
        self.neighboursIndexLists = None
        self.buildNeighboursIndexLists()
        self.buildNeighboursCoordLists()

    def getNbVoxels(self):
        return self.nbVoxels

    def getTargetAxesNames(self):
        return ['axial', 'coronal', 'sagittal']  # TODO make less static

    def getMapping(self):
        return self.mapping

    def getRoiMask(self):
        """
        Return a binary or n-ary 3D mask which has the shape of the target data
        """
        return NP.ones(self.shape, dtype=int)  # every position marked as 1

    def getNdArrayMask(self):
        return self.ndarrayMask

    def getCoord(self, index):
        return self.mapping[index]

    def getIndex(self, coord):
        return self.lattice[coord]

    def getNeighboursCoords(self, idvoxel):
        if self.neighboursCoordLists != None:
            return self.neighboursCoordLists[idvoxel]
        else:
            return [self.mapping[i]
                    for i in self.getNeighboursIndexes(idvoxel)]

    def getClosestNeighboursIndexes(self, idVoxel):
        cv = self.correctedMask[idVoxel]
        neighbours = []
        # translate center of neighbours mask to voxel position :
        neighboursTmp = self.neighbourMask + cv
        for cn in neighboursTmp:
            if NP.bitwise_and(cn >= 0, cn < self.shape).all():
                neighbours.append(cn.tolist())

        return [self.lattice[n[0], n[1], n[2]] for n in neighbours
                if self.lattice[n[0], n[1], n[2]] != -1]

    def getNeighboursIndexes(self, idVoxel):
        if self.neighboursIndexLists != None:
            return self.neighboursIndexLists[idVoxel]
        else:
            neighbours = set([idVoxel])
            for d in xrange(self.depth):
                newNeighbours = []
                for i in neighbours:
                    newNeighbours.extend(self.getClosestNeighboursIndexes(i))
                neighbours.update(newNeighbours)
            neighbours.remove(idVoxel)
            return list(neighbours)

    def buildNeighboursIndexLists(self):
        neighboursListsTmp = range(self.nbVoxels)
        for i in xrange(self.nbVoxels):
            neighboursListsTmp[i] = self.getNeighboursIndexes(i)

        self.neighboursIndexLists = neighboursListsTmp

    def buildNeighboursCoordLists(self):
        neighboursListsTmp = range(self.nbVoxels)
        for i in xrange(self.nbVoxels):
            neighboursListsTmp[i] = self.getNeighboursCoords(i)
        self.neighboursCoordLists = neighboursListsTmp

    def getNeighboursIndexLists(self):
        return self.neighboursIndexLists

    def getNeighboursCoordLists(self):
        return self.neighboursCoordLists

    def getNbCliques(self):
        if not hasattr(self, 'nbCliques') or self.nbCliques is None:
            self.nbCliques = sum([len(x)
                                  for x in self.neighboursIndexLists]) / 2.
        return self.nbCliques

    def mapVoxData(self, data, fillValue=0):
        if len(data.shape) == 1:
            mappedData = NP.zeros(self.shape, dtype=data.dtype) + fillValue
            mappedData[self.ndarrayMask] = data
        elif len(data.shape) == 2:  # assume (time, nbVox)
            sh = tuple([data.shape[0]]) + self.shape
            mappedData = NP.zeros(sh, dtype=data.dtype) + fillValue
            m = self.ndarrayCorMask
            mappedData[:, m[0], m[1], m[2]] = data

        return mappedData

    @staticmethod
    def createFromGUI(GUIobject):
        """
        Creates the actual object based on the parameters
        """
        params = {}
        params[RegularLatticeMapping.P_XSIZE] = getattr(
            GUIobject, RegularLatticeMapping.P_XSIZE)
        params[RegularLatticeMapping.P_YSIZE] = getattr(
            GUIobject, RegularLatticeMapping.P_YSIZE)
        params[RegularLatticeMapping.P_ZSIZE] = getattr(
            GUIobject, RegularLatticeMapping.P_ZSIZE)
        params[RegularLatticeMapping.P_ORDER] = getattr(
            GUIobject, RegularLatticeMapping.P_ORDER)
        params[RegularLatticeMapping.P_DEPTH] = getattr(
            GUIobject, RegularLatticeMapping.P_DEPTH)
        return RegularLatticeMapping(**params)


##################
# Spatial Fields #
##################


class StateField:
    """
    Class handling a field of states : a set of integers (ie labels) whose
    ranks can be spatially mapped to 3D coordinates. Each label refers to a
    class wich is identified by an ID and a name.
    """

    def __init__(self, classNames, spConf, initProps=None):
        """
        Initialise a StateField object.
        """
        self.classNames = classNames
        self.classIds = {}  # will map each class name to an integer class ID.
        ic = 0
        for cn in classNames:
            self.classIds[cn] = ic
            ic += 1

        self.nbClasses = len(self.classNames)
        self.spatialConfig = spConf
        self.initProportions = initProps
        self.size = spConf.getNbVoxels()
        self.data = NP.zeros(self.size, dtype=int)  # will store label values
        self.generate()  # generate label values
        self.classCounts = range(self.nbClasses)
        self.updateClassCounts()

    def getSize(self):
        """
        Return to size of the field.
        """
        return self.size

    def generate(self):
        """
        Generate values for every states. By default : if initProportions is set,
        generate values according to it. State values will be ordered by class ID
        """
        if self.initProportions != None:
            self.data = NP.array([], dtype=int)
            # Init labels according to init proportions :
            for ic in xrange(self.nbClasses - 1):
                cs = NP.round(self.size * self.initProportions[ic])
                self.data = NP.hstack((self.data, NP.zeros(cs) + ic))
            # Calculate last class size from the sum of the others :
            cs = self.size - len(self.data)
            self.data = NP.hstack(
                (self.data, NP.zeros(cs) + self.nbClasses - 1))
            # Update last proportions accordingly :
            ps = cs / (self.size + 0.)
            if len(self.initProportions) == self.nbClasses:
                self.initProportions[-1] = ps
            elif len(self.initProportions) == self.nbClasses - 1:
                self.initProportions.append(ps)
            else:
                # Error !
                pass

    def getNbClasses(self):
        return self.nbClasses

    def updateClassCounts(self):
        """
        Compute the size of every classes.
        """
        for ic in xrange(self.nbClasses):
            self.classCounts[ic] = (self.data == ic).sum()

    def randomize(self):
        """
        Randomize state values with a ramdom permutation.
        """
        self.data = NP.random.permutation(self.data)

    def getFieldValues(self):
        """
        Return all field values.
        """
        return self.data

    def getMappedFieldValues(self):
        return self.spatialConfig.mapVoxData(self.data)

    def setFieldValues0(self, values, mask=None):
        """
        Copy the content of 'values' to state values masked by 'mask'.
        """
        if mask == None:
            assert len(values) == len(self.data)
            self.data[:] = values[:]
        else:
            self.data[mask] = values[:]

    def setFieldValues(self, values, mask=None):
        """
        Copy the content of 'values' to state values masked by 'mask'.
        """
        if mask == None:
            assert len(values) == len(self.data)
            self.data[:] = values[:]
        else:
            self.data[mask] = values[mask]

    def getClassName(self, classId):
        """
        Return the class name corresponding to the integer 'classId'
        """
        assert type(classId) == int
        if classId < self.nbClasses and classId >= 0:
            return self.classNames[classId]
        else:
            return None

    def getClassNames(self):
        """
        Return all the class names
        """
        return self.classNames

    def getClassId(self, className):
        """
        Return the class id corresponding to the string 'className'
        """
        assert type(className) == str
        return self.classIds[className]


class PottsField(StateField):
    pass


def maskToMapping(m):
    return NP.array(NP.where(m != 0)).transpose()


def mask_to_coords(m):
    return NP.array(NP.where(m != 0)).transpose()
